from __future__ import annotations

import abc
import ast
import collections
import copy
import fnmatch
import gzip
import io
import logging
import pickle
import re
import typing
import weakref

from defusedxml import minidom

from hippolyzer.lib.base import serialization as se, llsd
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.datatypes import TaggedUnion, UUID, TupleCoord
from hippolyzer.lib.base.helpers import bytes_escape
from hippolyzer.lib.base.message.message_formatting import HumanMessageSerializer
from hippolyzer.lib.proxy.message_filter import MetaFieldSpecifier, compile_filter, BaseFilterNode, MessageFilterNode, \
    EnumFieldSpecifier, MatchResult
from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
from hippolyzer.lib.proxy.caps import CapType, SerializedCapData

if typing.TYPE_CHECKING:
    from hippolyzer.lib.proxy.region import ProxiedRegion
    from hippolyzer.lib.proxy.sessions import Session

LOG = logging.getLogger(__name__)


class BaseMessageLogger:
    paused: bool

    def log_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        if self.paused:
            return False
        return self.add_log_entry(LLUDPMessageLogEntry(message, region, session))

    def log_http_response(self, flow: HippoHTTPFlow):
        if self.paused:
            return False
        # These are huge, let's not log them for now.
        if flow.cap_data and flow.cap_data.asset_server_cap:
            return False
        return self.add_log_entry(HTTPMessageLogEntry(flow))

    def log_eq_event(self, session: Session, region: ProxiedRegion, event: dict):
        if self.paused:
            return False
        return self.add_log_entry(EQMessageLogEntry(event, region, session))

    @abc.abstractmethod
    def add_log_entry(self, entry: AbstractMessageLogEntry):
        pass


class FilteringMessageLogger(BaseMessageLogger):
    def __init__(self, maxlen=2000):
        BaseMessageLogger.__init__(self)
        self._raw_entries = collections.deque(maxlen=maxlen)
        self._filtered_entries: typing.List[AbstractMessageLogEntry] = []
        self.paused = False
        self.filter: BaseFilterNode = compile_filter("")

    def __iter__(self) -> typing.Iterator[AbstractMessageLogEntry]:
        return iter(self._filtered_entries)

    def set_filter(self, filter_str: str):
        self.filter = compile_filter(filter_str)
        self._begin_reset()
        # Keep any entries that've aged out of the raw entries list that
        # match the new filter
        self._filtered_entries = [
            m for m in self._filtered_entries if
            m not in self._raw_entries and self.filter.match(m)
        ]
        self._filtered_entries.extend((m for m in self._raw_entries if self.filter.match(m)))
        self._end_reset()

    def set_paused(self, paused: bool):
        self.paused = paused

    # Hooks that Qt models will want to implement
    def _begin_insert(self, insert_idx: int):
        pass

    def _end_insert(self):
        pass

    def _begin_reset(self):
        pass

    def _end_reset(self):
        pass

    def add_log_entry(self, entry: AbstractMessageLogEntry):
        try:
            # Paused, throw it away.
            if self.paused:
                return False
            self._raw_entries.append(entry)
            if self.filter.match(entry):
                next_idx = len(self._filtered_entries)
                self._begin_insert(next_idx)
                self._filtered_entries.append(entry)
                self._end_insert()
                return True
        except Exception:
            LOG.exception("Failed to filter queued message")
        return False

    def clear(self):
        self._begin_reset()
        self._filtered_entries.clear()
        self._raw_entries.clear()
        self._end_reset()


class WrappingMessageLogger(BaseMessageLogger):
    def __init__(self):
        self.loggers: typing.List[BaseMessageLogger] = []

    @property
    def paused(self):
        return all(x.paused for x in self.loggers)

    def add_log_entry(self, entry: AbstractMessageLogEntry):
        logged = False
        for logger in self.loggers:
            if logger.add_log_entry(entry):
                logged = True
        # At least one logger ended up keeping the message around, so let's
        # cache the summary before we freeze the message.
        if logged:
            entry.cache_summary()
        entry.freeze()


class AbstractMessageLogEntry(abc.ABC):
    region: typing.Optional[ProxiedRegion]
    session: typing.Optional[Session]
    name: str
    type: str

    __slots__ = ["_region", "_session", "_region_name", "_agent_id", "_summary", "meta"]

    def __init__(self, region: typing.Optional[ProxiedRegion], session: typing.Optional[Session]):
        if region and not isinstance(region, weakref.ReferenceType):
            region = weakref.ref(region)
        if session and not isinstance(session, weakref.ReferenceType):
            session = weakref.ref(session)

        self._region: typing.Optional[weakref.ReferenceType] = region
        self._session: typing.Optional[weakref.ReferenceType] = session
        self._region_name = None
        self._agent_id = None
        self._summary = None
        if self.region:
            self._region_name = self.region.name
        if self.session:
            self._agent_id = self.session.agent_id

        agent_obj = None
        if self.region is not None:
            agent_obj = self.region.objects.lookup_fullid(self.agent_id)
        self.meta = {
            "RegionName": self.region_name,
            "AgentID": self.agent_id,
            "SessionID": self.session.id if self.session else None,
            "AgentLocal": agent_obj.LocalID if agent_obj is not None else None,
            "Method": self.method,
            "Type": self.type,
            "SelectedLocal": self._current_selected_local(),
            "SelectedFull": self._current_selected_full(),
        }

    def to_dict(self) -> dict:
        meta = self.meta.copy()

        def _dehydrate_meta_uuid(key: str):
            if meta[key]:
                meta[key] = str(meta[key])

        _dehydrate_meta_uuid("AgentID")
        _dehydrate_meta_uuid("SelectedFull")
        _dehydrate_meta_uuid("SessionID")

        return {
            "type": self.type,
            "region_name": self.region_name,
            "agent_id": str(self.agent_id) if self.agent_id is not None else None,
            "summary": self.summary,
            "meta": meta,
        }

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, val: dict):
        pass

    def apply_dict(self, val: dict) -> None:
        self._region_name = val['region_name']
        self._agent_id = UUID(val['agent_id']) if val['agent_id'] else None
        self._summary = val['summary']
        meta = val['meta'].copy()

        def _hydrate_meta_uuid(key: str):
            if meta[key]:
                meta[key] = UUID(meta[key])

        _hydrate_meta_uuid("AgentID")
        _hydrate_meta_uuid("SelectedFull")
        _hydrate_meta_uuid("SessionID")
        self.meta.update(meta)

    def freeze(self):
        pass

    def cache_summary(self):
        self._summary = self.summary

    def _current_selected_local(self):
        if self.session:
            return self.session.selected.object_local
        return None

    def _current_selected_full(self):
        selected_local = self._current_selected_local()
        if selected_local is None or self.region is None:
            return None
        obj = self.region.objects.lookup_localid(selected_local)
        return obj and obj.FullID

    def _get_meta(self, name: str) -> typing.Any:
        # Slight difference in semantics. Filters are meant to return the same
        # thing no matter when they're run, so SelectedLocal and friends resolve
        # to the selected items _at the time the message was logged_. To handle
        # the case where we want to match on the selected object at the time the
        # filter is evaluated, we resolve these here.
        if name == "CurrentSelectedLocal":
            return self._current_selected_local()
        elif name == "CurrentSelectedFull":
            return self._current_selected_full()
        return self.meta.get(name)

    @property
    def region(self) -> typing.Optional[ProxiedRegion]:
        if self._region:
            return self._region()
        return None

    @property
    def session(self) -> typing.Optional[Session]:
        if self._session:
            return self._session()
        return None

    @property
    def region_name(self) -> str:
        region = self.region
        if region:
            self._region_name = region.name
            return self._region_name
        # Region may die after a message is logged, need to keep this around.
        if self._region_name:
            return self._region_name

        return ""

    @property
    def agent_id(self) -> typing.Optional[UUID]:
        if self._agent_id:
            return self._agent_id

        session = self.session
        if session:
            self._agent_id = session.agent_id
            return self._agent_id
        return None

    @property
    def host(self) -> str:
        region_name = self.region_name
        if not region_name:
            return ""
        session_str = ""
        agent_id = self.agent_id
        if agent_id:
            session_str = f" ({agent_id})"
        return region_name + session_str

    def request(self, beautify=False, replacements=None):
        return None

    def response(self, beautify=False):
        return None

    def _packet_root_matches(self, pattern):
        if fnmatch.fnmatchcase(self.name, pattern):
            return True
        if fnmatch.fnmatchcase(self.type, pattern):
            return True
        return False

    def _val_matches(self, operator, val, expected):
        if isinstance(expected, MetaFieldSpecifier):
            if len(expected) != 1:
                raise ValueError(f"Can only support single-level Meta specifiers, not {expected!r}")
            expected = self._get_meta(str(expected[0]))
            if not isinstance(expected, (int, float, bytes, str, type(None), tuple)):
                if callable(expected):
                    expected = expected()
                else:
                    expected = str(expected)
        elif isinstance(expected, EnumFieldSpecifier):
            # Local import so we get a fresh copy of the templates module
            from hippolyzer.lib.proxy import templates
            enum_cls = getattr(templates, expected.enum_name)
            expected = enum_cls[expected.field_name]
        elif expected is not None:
            # Unbox the expected value
            expected = expected.value
        if not isinstance(val, (int, float, bytes, str, type(None), tuple, TupleCoord)):
            val = str(val)

        if not operator:
            return bool(val)
        elif operator == "==":
            return val == expected
        elif operator == "!=":
            return val != expected
        elif operator == "^=":
            if val is None:
                return False
            return val.startswith(expected)
        elif operator == "$=":
            if val is None:
                return False
            return val.endswith(expected)
        elif operator == "~=":
            if val is None:
                return False
            return expected in val
        elif operator == "<":
            return val < expected
        elif operator == "<=":
            return val <= expected
        elif operator == ">":
            return val > expected
        elif operator == ">=":
            return val >= expected
        elif operator == "&":
            return val & expected
        else:
            raise ValueError(f"Unexpected operator {operator!r}")

    def _base_matches(self, matcher: "MessageFilterNode") -> typing.Optional[bool]:
        if len(matcher.selector) == 1:
            # Comparison operators would make no sense here
            if matcher.value or matcher.operator:
                return False
            return self._packet_root_matches(matcher.selector[0])
        if matcher.selector[0] == "Meta":
            if len(matcher.selector) == 2:
                return self._val_matches(matcher.operator, self._get_meta(matcher.selector[1]), matcher.value)
            elif len(matcher.selector) == 3:
                meta_dict = self._get_meta(matcher.selector[1])
                if not meta_dict or not hasattr(meta_dict, 'get'):
                    return False
                return self._val_matches(matcher.operator, meta_dict.get(matcher.selector[2]), matcher.value)
        return None

    def matches(self, matcher: "MessageFilterNode", short_circuit=True) -> "MatchResult":
        return MatchResult(self._base_matches(matcher) or False, [])

    @property
    def seq(self):
        return ""

    @property
    def method(self):
        return ""

    @property
    def summary(self):
        return ""

    @staticmethod
    def _format_llsd(parsed):
        xmlified = llsd.format_pretty_xml(parsed)
        # dedent <key> by 1 for easier visual scanning
        xmlified = re.sub(rb" <key>", b"<key>", xmlified)
        return xmlified.decode("utf8", errors="replace")

    @staticmethod
    def _format_xml(content):
        beautified = minidom.parseString(content).toprettyxml(indent="  ")
        # kill blank lines. will break cdata sections. meh.
        beautified = re.sub(r'\n\s*\n', '\n', beautified, flags=re.MULTILINE)
        return re.sub(r'<(\w+)>\s*</\1>', r'<\1></\1>',
                      beautified, flags=re.MULTILINE)


class HTTPMessageLogEntry(AbstractMessageLogEntry):
    __slots__ = ["flow"]

    def __init__(self, flow: HippoHTTPFlow):
        self.flow: HippoHTTPFlow = flow
        cap_data = self.flow.cap_data
        region = cap_data and cap_data.region
        session = cap_data and cap_data.session

        super().__init__(region, session)
        # This was a request the proxy made through itself
        self.meta["Synthetic"] = flow.request_injected

    @property
    def type(self):
        return "HTTP"

    @property
    def name(self):
        cap_data = self.flow.cap_data
        name = cap_data and cap_data.cap_name
        if name:
            return name
        return self.flow.request.url

    @property
    def method(self):
        return self.flow.request.method

    def _format_http_message(self, want_request, beautify):
        message = self.flow.request if want_request else self.flow.response
        method = self.flow.request.method
        buf = io.StringIO()
        cap_data = self.flow.cap_data
        cap_name = cap_data and cap_data.cap_name
        base_url = cap_name and cap_data.base_url
        temporary_cap = cap_data and cap_data.type == CapType.TEMPORARY
        beautify_url = (beautify and base_url and cap_name
                        and not temporary_cap and self.session and want_request)
        if want_request:
            buf.write(message.method)
            buf.write(" ")
            if beautify_url:
                buf.write(f"[[{cap_name}]]{message.url[len(base_url):]}")
            else:
                buf.write(message.url)
            buf.write(" ")
            buf.write(message.http_version)
        else:
            buf.write(message.http_version)
            buf.write(" ")
            buf.write(str(message.status_code))
            buf.write(" ")
            buf.write(message.reason)
        buf.write("\r\n")
        if beautify_url:
            buf.write("# ")
            buf.write(message.url)
            buf.write("\r\n")

        headers = copy.deepcopy(message.headers)
        for key in tuple(headers.keys()):
            if key.lower().startswith("x-hippo-"):
                LOG.warning(f"Internal header {key!r} leaked out?")
                # If this header actually came from somewhere untrusted, we can't
                # include it. It may change the meaning of the message when replayed.
                headers[f"X-Untrusted-{key}"] = headers[key]
                headers.pop(key)
        beautified = None
        if beautify and message.content:
            try:
                serializer = se.HTTP_SERIALIZERS.get(cap_name)
                if serializer:
                    if want_request:
                        beautified = serializer.deserialize_req_body(method, message.content)
                    else:
                        beautified = serializer.deserialize_resp_body(method, message.content)

                    if beautified is se.UNSERIALIZABLE:
                        beautified = None
                    else:
                        beautified = self._format_llsd(beautified)
                        headers["X-Hippo-Beautify"] = "1"

                if not beautified:
                    content_type = self._guess_content_type(message)
                    if content_type.startswith("application/llsd"):
                        try:
                            beautified = self._format_llsd(llsd.parse(message.content))
                        except llsd.LLSDParseError:
                            # Sometimes LL sends plain XML with a Content-Type of application/llsd+xml.
                            # Try to detect that case and work around it
                            if content_type == "application/llsd+xml" and message.content.startswith(b'<'):
                                beautified = self._format_xml(message.content)
                            else:
                                raise
                    elif any(content_type.startswith(x) for x in ("application/xml", "text/xml")):
                        beautified = self._format_xml(message.content)
            except:
                LOG.exception("Failed to beautify message")

        message_body = beautified or message.content
        if isinstance(message_body, bytes):
            try:
                decoded = message.text
                # Valid in many codecs, but unprintable.
                if "\x00" in decoded:
                    raise ValueError("Embedded null")
                message_body = decoded
            except (UnicodeError, ValueError):
                # non-printable characters, return the escaped version.
                headers["X-Hippo-Escaped-Body"] = "1"
                message_body = bytes_escape(message_body).decode("utf8")

        buf.write(bytes(headers).decode("utf8", errors="replace"))
        buf.write("\r\n")

        buf.write(message_body or "")
        return buf.getvalue()

    def request(self, beautify=False, replacements=None):
        return self._format_http_message(want_request=True, beautify=beautify)

    def response(self, beautify=False):
        return self._format_http_message(want_request=False, beautify=beautify)

    @property
    def summary(self):
        if self._summary is not None:
            return self._summary
        msg = self.flow.response
        self._summary = f"{msg.status_code}: "
        if not msg.content:
            return self._summary
        if len(msg.content) > 1000000:
            self._summary += "[too large...]"
            return self._summary
        content_type = self._guess_content_type(msg)
        if content_type.startswith("application/llsd"):
            notation = llsd.format_notation(llsd.parse(msg.content))
            self._summary += notation.decode("utf8")[:500]
        return self._summary

    def _guess_content_type(self, message):
        # SL's login service lies and says that its XML-RPC response is LLSD+XML.
        # It is not, and it blows up the parser. It's been broken ever since the
        # login rewrite and a fix is likely not forthcoming. I'm sick of seeing
        # the traceback, so just hack around it.
        if self.name == "LoginRequest":
            return "application/xml"
        content_type = message.headers.get("Content-Type", "")
        if not message.content or content_type.startswith("application/llsd"):
            return content_type
        # Sometimes gets sent with `text/plain` or `text/html`. Cool.
        if message.content.startswith(rb'<?xml version="1.0" ?><llsd>'):
            return "application/llsd+xml"
        if message.content.startswith(rb'<llsd>'):
            return "application/llsd+xml"
        if message.content.startswith(rb'<?xml '):
            return "application/xml"
        return content_type

    def _get_meta(self, name: str) -> typing.Any:
        lower_name = name.lower()
        if lower_name == "url":
            return self.flow.request.url
        elif lower_name == "reqheaders":
            return self.flow.request.headers
        elif lower_name == "respheaders":
            return self.flow.response.headers
        elif lower_name == "host":
            return self.flow.request.host.lower()
        elif lower_name == "status":
            return self.flow.response.status_code
        return super()._get_meta(name)

    def to_dict(self):
        val = super().to_dict()
        val['flow'] = self.flow.get_state()
        cap_data = val['flow'].get('metadata', {}).get('cap_data_ser')
        if cap_data is not None:
            # Have to convert this from a namedtuple to a dict to make
            # it importable
            cap_dict = cap_data._asdict()  # noqa
            val['flow']['metadata']['cap_data_ser'] = cap_dict
        return val

    @classmethod
    def from_dict(cls, val: dict):
        cap_data = val['flow'].get('metadata', {}).get('cap_data_ser')
        if cap_data:
            val['flow']['metadata']['cap_data_ser'] = SerializedCapData(**cap_data)
        ev = cls(HippoHTTPFlow.from_state(val['flow'], None))
        ev.apply_dict(val)
        return ev


class EQMessageLogEntry(AbstractMessageLogEntry):
    __slots__ = ["event"]

    def __init__(self, event, region, session):
        super().__init__(region, session)
        self.event = event

    @property
    def type(self):
        return "EQ"

    def request(self, beautify=False, replacements=None):
        return f'EQ {self.event["message"]}\n\n{self._format_llsd(self.event["body"])}'

    @property
    def name(self):
        return self.event["message"]

    @property
    def summary(self):
        if self._summary is not None:
            return self._summary
        self._summary = ""
        self._summary = llsd.format_notation(self.event["body"]).decode("utf8")[:500]
        return self._summary

    def to_dict(self) -> dict:
        val = super().to_dict()
        val['event'] = llsd.format_notation(self.event)
        return val

    @classmethod
    def from_dict(cls, val: dict):
        ev = cls(llsd.parse_notation(val['event']), None, None)
        ev.apply_dict(val)
        return ev


class LLUDPMessageLogEntry(AbstractMessageLogEntry):
    __slots__ = ["_message", "_name", "_direction", "_frozen_message", "_seq", "_deserializer"]

    def __init__(self, message: Message, region, session):
        self._message: Message = message
        self._deserializer = None
        self._name = message.name
        self._direction = message.direction
        self._frozen_message: typing.Optional[bytes] = None
        self._seq = message.packet_id
        super().__init__(region, session)

    _MESSAGE_META_ATTRS = {
        "Synthetic", "Dropped", "Extra", "Resent", "Zerocoded", "Acks", "Reliable",
    }

    def _get_meta(self, name: str):
        # These may change between when the message is logged and when we
        # actually filter on it, since logging happens before addons.
        msg = self.message
        if name in self._MESSAGE_META_ATTRS:
            return getattr(msg, name.lower(), None)
        msg_meta = getattr(msg, "meta", None)
        if msg_meta is not None:
            if name in msg_meta:
                return msg_meta[name]
        return super()._get_meta(name)

    @property
    def message(self) -> Message:
        if self._message:
            return self._message
        elif self._frozen_message:
            message = pickle.loads(self._frozen_message)
            message.deserializer = self._deserializer
            return message
        else:
            raise ValueError("Didn't have a fresh or frozen message somehow")

    def freeze(self):
        message = self.message
        message.invalidate_caches()
        # These are expensive to keep around. pickle them and un-pickle on
        # an as-needed basis.
        self._deserializer = self.message.deserializer
        message.deserializer = None
        try:
            self._frozen_message = pickle.dumps(self._message, protocol=pickle.HIGHEST_PROTOCOL)
        finally:
            message.deserializer = self._deserializer
        self._message = None

    @property
    def type(self):
        return "LLUDP"

    @property
    def name(self):
        if self._message:
            self._name = self._message.name
        return self._name

    @property
    def method(self):
        if self._message:
            self._direction = self._message.direction
        return self._direction.name if self._direction is not None else ""

    def request(self, beautify=False, replacements=None):
        return HumanMessageSerializer.to_human_string(self.message, replacements, beautify)

    def matches(self, matcher, short_circuit=True) -> "MatchResult":
        base_matched = self._base_matches(matcher)
        if base_matched is not None:
            return MatchResult(base_matched, [])

        if not self._packet_root_matches(matcher.selector[0]):
            return MatchResult(False, [])

        message = self.message

        selector_len = len(matcher.selector)
        # name, block_name, var_name(, subfield_name)?
        if selector_len not in (3, 4):
            return MatchResult(False, [])
        found_field_keys = []
        for block_name in message.blocks:
            if not fnmatch.fnmatchcase(block_name, matcher.selector[1]):
                continue
            for block_num, block in enumerate(message[block_name]):
                for var_name in block.vars.keys():
                    if not fnmatch.fnmatchcase(var_name, matcher.selector[2]):
                        continue
                    # So we know where the match happened
                    field_key = (message.name, block_name, block_num, var_name)
                    if selector_len == 3:
                        # We're just matching on the var existing, not having any particular value
                        if matcher.value is None:
                            found_field_keys.append(field_key)
                        elif self._val_matches(matcher.operator, block[var_name], matcher.value):
                            found_field_keys.append(field_key)
                    # Need to invoke a special unpacker
                    elif selector_len == 4:
                        try:
                            deserialized = block.deserialize_var(var_name)
                        except KeyError:
                            continue
                        # Discard the tag if this is a tagged union, we only want the value
                        if isinstance(deserialized, TaggedUnion):
                            deserialized = deserialized.value
                        if not isinstance(deserialized, dict):
                            continue
                        for key in deserialized.keys():
                            if fnmatch.fnmatchcase(str(key), matcher.selector[3]):
                                if matcher.value is None:
                                    # Short-circuiting checking individual subfields is fine since
                                    # we only highlight fields anyway.
                                    found_field_keys.append(field_key)
                                    break
                                elif self._val_matches(matcher.operator, deserialized[key], matcher.value):
                                    found_field_keys.append(field_key)
                                    break

                    if short_circuit and found_field_keys:
                        return MatchResult(True, found_field_keys)
        return MatchResult(bool(found_field_keys), found_field_keys)

    @property
    def summary(self):
        if self._summary is None:
            self._summary = self.message.to_summary()[:500]
        return self._summary

    @property
    def seq(self):
        if self._message:
            self._seq = self._message.packet_id
        return self._seq

    def to_dict(self):
        val = super().to_dict()
        val['message'] = llsd.format_notation(self.message.to_dict(extended=True))
        return val

    @classmethod
    def from_dict(cls, val: dict):
        ev = cls(Message.from_dict(llsd.parse_notation(val['message'])), None, None)
        ev.apply_dict(val)
        return ev


def export_log_entries(entries: typing.Iterable[AbstractMessageLogEntry]) -> bytes:
    return gzip.compress(repr([e.to_dict() for e in entries]).encode("utf8"))


_TYPE_CLASSES = {
    "HTTP": HTTPMessageLogEntry,
    "LLUDP": LLUDPMessageLogEntry,
    "EQ": EQMessageLogEntry,
}


def import_log_entries(data: bytes) -> typing.List[AbstractMessageLogEntry]:
    entries = ast.literal_eval(gzip.decompress(data).decode("utf8"))
    return [_TYPE_CLASSES[e['type']].from_dict(e) for e in entries]



"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import asyncio
import contextlib
import logging
from typing import *

from hippolyzer.lib.base.events import Event

LOG = logging.getLogger(__name__)
_T = TypeVar("_T")
_K = TypeVar("_K", bound=Hashable)
MESSAGE_HANDLER = Callable[[_T], Any]
PREDICATE = Callable[[_T], bool]
MESSAGE_NAMES = Iterable[_K]


class MessageHandler(Generic[_T, _K]):
    def __init__(self, take_by_default: bool = True):
        self.handlers: Dict[_K, Event] = {}
        self.take_by_default = take_by_default

    def register(self, message_name: _K) -> Event:
        LOG.debug('Creating a monitor for %s' % message_name)
        return self.handlers.setdefault(message_name, Event())

    def subscribe(self, message_name: _K, handler: MESSAGE_HANDLER) -> Event:
        notifier = self.register(message_name)
        notifier.subscribe(handler)
        return notifier

    def _subscribe_all(self, message_names: MESSAGE_NAMES, handler: MESSAGE_HANDLER,
                       predicate: Optional[PREDICATE] = None) -> List[Event]:
        notifiers = [self.register(name) for name in message_names]
        for n in notifiers:
            n.subscribe(handler, predicate=predicate)
        return notifiers

    @contextlib.contextmanager
    def subscribe_async(self, message_names: MESSAGE_NAMES, predicate: Optional[PREDICATE] = None,
                        take: Optional[bool] = None) -> ContextManager[Callable[[], Awaitable[_T]]]:
        """
        Subscribe to a set of message matching predicate while within a block

        Defaults to taking the message out of the usual flow, and any matching messages will
        not be automatically be forwarded through to the client, allowing the subscriber coroutine
        time to modify or drop the message. Taken messages must be manually sent to the client by
        subscribers if desired.

        If a subscriber is just an observer that will never drop or modify a message, take=False
        may be used and messages will be sent as usual.
        """
        if take is None:
            take = self.take_by_default
        msg_queue = asyncio.Queue()

        def _handler_wrapper(message: _T):
            # Consider this message owned by one of the async handlers, drop it
            if take:
                message = message.take()
            msg_queue.put_nowait(message)

        notifiers = self._subscribe_all(message_names, _handler_wrapper, predicate=predicate)

        async def _get_wrapper():
            msg = await msg_queue.get()
            # Consumption is completion
            msg_queue.task_done()
            return msg

        try:
            yield _get_wrapper
        finally:
            for n in notifiers:
                n.unsubscribe(_handler_wrapper)

    def wait_for(self, message_names: MESSAGE_NAMES, predicate: Optional[PREDICATE] = None,
                 timeout: Optional[float] = None, take: Optional[bool] = None) -> Awaitable[_T]:
        """
        Wait for a single instance one of message_names matching predicate

        Any packets matching predicate will be considered owned by the caller and will be
        automatically dropped unless `take=False`. This should not be used if waiting for a
        sequence of packets, since multiple packets may come in after the future has already
        been marked completed, causing some to be missed.
        """
        if take is None:
            take = self.take_by_default
        notifiers = [self.register(name) for name in message_names]

        loop = asyncio.get_event_loop_policy().get_event_loop()
        fut = loop.create_future()
        timeout_task = None

        async def _canceller():
            await asyncio.sleep(timeout)
            if not fut.done():
                fut.set_exception(asyncio.exceptions.TimeoutError("Timed out waiting for packet"))
            for n in notifiers:
                n.unsubscribe(_handler)

        if timeout:
            timeout_task = asyncio.create_task(_canceller())

        def _handler(message: _T):
            if timeout_task:
                timeout_task.cancel()
            # Whatever was awaiting this future now owns this message
            if take:
                message = message.take()
            if not fut.done():
                fut.set_result(message)
            # Make sure to unregister this handler for all message types
            for n in notifiers:
                n.unsubscribe(_handler)

        for notifier in notifiers:
            notifier.subscribe(_handler, predicate=predicate)
        return fut

    def is_handled(self, message_name: _K):
        return message_name in self.handlers

    def handle(self, message: _T):
        self._handle_type(message.name, message)
        # Always try to call wildcard handlers
        self._handle_type('*', message)

    def _handle_type(self, name: _K, message: _T):
        handler = self.handlers.get(name)
        if not handler:
            return

        if len(handler) > 0:
            LOG.debug('Handling message : %s' % name)
            handler(message)



from __future__ import annotations

import asyncio
import functools
import threading
from typing import *

from hippolyzer.lib.proxy.task_scheduler import TaskLifeScope

if TYPE_CHECKING:
    from hippolyzer.lib.proxy.region import ProxiedRegion
    from hippolyzer.lib.proxy.sessions import Session
    WrappedCommandCallable = Callable[[Session, ProxiedRegion, str], Coroutine]


class Parameter(NamedTuple):
    parser: Callable[[str], Any] = str
    # None makes this greedy
    sep: Optional[str] = " "
    optional: bool = False


class CommandDetails(NamedTuple):
    name: str
    params: Dict[str, Parameter]
    lifetime: Optional[TaskLifeScope] = None


def parse_bool(val: str) -> bool:
    return val.lower() in ('on', 'true', '1', '1.0', 'yes')


def handle_command(command_name: Optional[str] = None, /, *, lifetime: Optional[TaskLifeScope] = None,
                   single_instance: bool = False, **params: Union[Parameter, callable]):
    """
    Register a coroutinefunction as a handler for a named command

    The function's name will be used if a name is not provided.
    """
    def _command_wrapper(func: Callable[..., Coroutine]):
        if not asyncio.iscoroutinefunction(func):
            raise ValueError(f"{func!r} is not async!")

        if single_instance:
            inner_func = func
            lock = threading.Lock()

            @functools.wraps(inner_func)
            async def func(self, *args, **kwargs):
                if lock.acquire(blocking=False):
                    try:
                        return await inner_func(self, *args, **kwargs)
                    finally:
                        lock.release()
                else:
                    raise RuntimeError(f"{inner_func.__name__} is already running")

        @functools.wraps(func)
        def _inner(self, session: Session, region: ProxiedRegion, message: str) -> Coroutine:
            param_vals = {}
            for param_name, param in params.items():
                # can do foo=str, bar=int
                if callable(param):
                    param = Parameter(parser=param)
                # Greedy, takes the rest of the message
                if param.sep is None:
                    param_val = message
                    message = ""
                else:
                    message = message.lstrip(param.sep)
                    if not message:
                        if not param.optional:
                            raise KeyError(f"Missing parameter {param_name}")
                        continue
                    param_val, _, message = message.partition(param.sep)  # type: ignore

                param_vals[param_name] = param.parser(param_val)

            return func(self, session, region, **param_vals)

        _inner.command = CommandDetails(
            name=command_name or func.__name__,
            lifetime=lifetime,
            params={k: v or Parameter() for (k, v) in params.items()},
        )
        return _inner
    return _command_wrapper



from __future__ import annotations

import collections
import datetime as dt
from typing import *

from mitmproxy import http

from hippolyzer.lib.base.datatypes import UUID
if TYPE_CHECKING:
    from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow


class AssetData(NamedTuple):
    data: bytes
    expires: Optional[dt.datetime]


class HTTPAssetRepo(collections.UserDict):
    data: Dict[UUID, AssetData]

    def create_asset(self, data, one_shot=False) -> UUID:
        asset_id = UUID.random()
        expires = None
        if one_shot:
            # If there are multiple concurrent sessions then we may see multiple
            # requests for an asset even if we only really need one per-client.
            # Deal with that by evicting after a short period instead of immediately
            # evicting on first request.
            expires = dt.datetime.now() + dt.timedelta(seconds=5)
        self.data[asset_id] = AssetData(data, expires)
        return asset_id

    def collect_garbage(self):
        for k in tuple(self.data.keys()):
            v = self.data[k]
            if not v.expires:
                continue
            if v.expires <= dt.datetime.now():
                del self.data[k]

    def try_serve_asset(self, flow: HippoHTTPFlow) -> bool:
        self.collect_garbage()

        if not flow.cap_data.asset_server_cap:
            return False

        asset_id = None
        for name, val in flow.request.query.items():
            if name.endswith("_id"):
                try:
                    asset_id = UUID(val)
                    break
                except ValueError:
                    pass

        if not asset_id or asset_id not in self.data:
            return False

        asset = self[asset_id]
        flow.response = http.Response.make(
            content=asset.data,
            headers={
                "Content-Type": "application/octet-stream",
            }
        )
        return True



from __future__ import annotations

import abc
import asyncio
import collections
import ctypes
import importlib.util
import inspect
import logging
import os
import pathlib
import sys
import textwrap
import time
from types import ModuleType
from typing import *

import outleap

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.helpers import get_mtime
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.network.transport import UDPPacket
from hippolyzer.lib.proxy import addon_ctx
from hippolyzer.lib.proxy.task_scheduler import TaskLifeScope, TaskScheduler

if TYPE_CHECKING:
    from hippolyzer.lib.proxy.commands import CommandDetails, WrappedCommandCallable
    from hippolyzer.lib.proxy.http_flow import HippoHTTPFlow
    from hippolyzer.lib.proxy.object_manager import Object
    from hippolyzer.lib.proxy.region import ProxiedRegion
    from hippolyzer.lib.proxy.sessions import Session, SessionManager

LOG = logging.getLogger(__name__)


class BaseInteractionManager:
    @abc.abstractmethod
    async def open_dir(self, caption: str = '', directory: str = '', filter_str: str = '') -> Optional[str]:
        pass

    @abc.abstractmethod
    async def open_files(self, caption: str = '', directory: str = '', filter_str: str = '') -> List[str]:
        pass

    @abc.abstractmethod
    async def open_file(self, caption: str = '', directory: str = '', filter_str: str = '') -> Optional[str]:
        pass

    @abc.abstractmethod
    async def save_file(self, caption: str = '', directory: str = '', filter_str: str = '',
                        default_suffix: str = '') -> Optional[str]:
        pass

    @abc.abstractmethod
    async def confirm(self, title: str, caption: str) -> bool:
        pass

    def main_window_handle(self) -> Any:
        return None


# Used to initialize a REPL environment with commonly desired helpers
REPL_INITIALIZER = r"""
from hippolyzer.lib.base.datatypes import *
from hippolyzer.lib.base.templates import *
from hippolyzer.lib.base.message.message import Block, Message, Direction
from hippolyzer.lib.proxy.addon_utils import send_chat, show_message
"""


class AddonManager:
    COMMAND_CHANNEL = 524

    BASE_ADDON_SPECS = []
    FILE_MTIMES: Dict[str, float] = {}
    HOTRELOAD_IMPORTERS: Dict[str, set] = collections.defaultdict(set)
    FRESH_ADDON_MODULES: Dict[str, Any] = {}
    SESSION_MANAGER: Optional[SessionManager] = None
    UI: Optional[BaseInteractionManager] = None
    LAST_RELOAD = None
    SCHEDULER = TaskScheduler()
    _SUBPROCESS: bool = False
    _REPL_TASK: Optional[asyncio.Task] = None
    _HOT_RELOADING_STACK: Set[str] = set()
    _SWALLOW_ADDON_EXCEPTIONS: bool = True

    @classmethod
    def init(cls, addon_script_paths, session_manager, addon_objects=None, subprocess=False,
             swallow_addon_exceptions=True):
        addon_objects = addon_objects or []
        cls.BASE_ADDON_SPECS.clear()
        cls.FRESH_ADDON_MODULES.clear()
        cls.FILE_MTIMES.clear()
        cls.LAST_RELOAD = None
        cls.SESSION_MANAGER = session_manager
        # Are we a child process from `multiprocessing`?
        cls._SUBPROCESS = subprocess
        cls._HOT_RELOADING_STACK.clear()
        cls._SWALLOW_ADDON_EXCEPTIONS = swallow_addon_exceptions
        for path in addon_script_paths:
            # defer reloading until we've collected them all
            cls.load_addon_from_path(path, reload=False)
        for addon in addon_objects:
            cls.FRESH_ADDON_MODULES[UUID.random()] = addon
        cls._reload_addons()

    @classmethod
    def shutdown(cls):
        to_pop = []
        for mod in cls.FRESH_ADDON_MODULES.values():
            to_pop.append(mod)
            cls._call_module_hooks(mod, "handle_unload", cls.SESSION_MANAGER)
        cls.SCHEDULER.shutdown()
        for mod in to_pop:
            if isinstance(mod, ModuleType):
                sys.modules.pop(mod.__name__, None)

    @classmethod
    def have_active_repl(cls):
        return bool(cls._REPL_TASK and not cls._REPL_TASK.done())

    @classmethod
    def spawn_repl(cls, _globals=None, _locals=None, skip_frames=1) -> asyncio.Task:
        """
        Spawn a non-blocking REPL in the context of the caller's stack frame

        Execution will continue as usual after the REPL starts. If blocking is desired,
        the caller can `await` the task or run the proxy within a more suitable
        debugger like `ipdb`.

        Any REPL active when the proxy closes will be automatically cleaned up.
        """
        if cls.have_active_repl():
            raise RuntimeError("Only one active REPL allowed at a time")

        # Function-level import because this is an optional requirement
        import ptpython.repl

        # Walk up the stack and pull locals / globals from there
        stack = inspect.stack()[skip_frames]
        if _globals is None:
            _globals = stack.frame.f_globals
        if _locals is None:
            _locals = stack.frame.f_locals

        init_globals = {}
        exec(REPL_INITIALIZER, init_globals, None)
        # We're modifying the globals of the caller, be careful of things we imported
        # for the REPL initializer clobber things that already exist in the caller's globals.
        # Making our own mutable copy of the globals dict, mutating that and then passing it
        # to embed() is not an option due to https://github.com/prompt-toolkit/ptpython/issues/279
        for global_name, global_val in init_globals.items():
            if global_name not in _globals:
                _globals[global_name] = global_val

        async def _wrapper():
            coro: Coroutine = ptpython.repl.embed(  # noqa: the type signature lies
                globals=_globals,
                locals=_locals,
                return_asyncio_coroutine=True,
                patch_stdout=True,
            )
            await coro
            # Merge our changes to locals and globals back into the real stack frame
            ctypes.pythonapi.PyFrame_LocalsToFast(
                ctypes.py_object(stack.frame), ctypes.c_int(0))

        task = cls.SCHEDULER.schedule_task(_wrapper())
        cls._REPL_TASK = task
        return task

    @classmethod
    def load_addon_from_path(cls, path, reload=False, raise_exceptions=True):
        path = pathlib.Path(path).absolute()
        mod_name = "hippolyzer.user_addon_%s" % path.stem
        cls.BASE_ADDON_SPECS.append(importlib.util.spec_from_file_location(mod_name, path))
        addon_dir = os.path.realpath(pathlib.Path(path).parent.absolute())

        if addon_dir not in sys.path:
            sys.path.append(addon_dir)

        # Addon list is dirty, force a reload
        cls.LAST_RELOAD = None
        if reload:
            cls._reload_addons(raise_exceptions=raise_exceptions)

    @classmethod
    def unload_addon_from_path(cls, path, reload=False):
        path = str(pathlib.Path(path).absolute())
        specs = [x for x in cls.BASE_ADDON_SPECS if x.origin == path]
        cls.BASE_ADDON_SPECS.remove(specs[0])
        cls.FILE_MTIMES.pop(path, None)
        old_mod = cls.FRESH_ADDON_MODULES.pop(specs[0].name, None)
        if old_mod:
            cls._unload_module(old_mod)
            sys.modules.pop(old_mod.__name__, None)
        if reload:
            cls._reload_addons()

    @classmethod
    def _check_hotreloads(cls):
        """Mark addons that rely on changed files for reloading"""
        for file_path, importers in cls.HOTRELOAD_IMPORTERS.items():
            mtime = get_mtime(file_path)
            if not mtime or mtime == cls.FILE_MTIMES.get(file_path, None):
                continue

            # Mark anything that imported this as dirty too, handling circular
            # dependencies.
            seen_importers = set()

            def _dirty_importers(cur_importers: Set[str]):
                for importer in cur_importers:
                    if importer in seen_importers:
                        continue
                    seen_importers.add(importer)
                    if importer in cls.FILE_MTIMES:
                        del cls.FILE_MTIMES[importer]
                    if importer in cls.HOTRELOAD_IMPORTERS:
                        _dirty_importers(cls.HOTRELOAD_IMPORTERS[importer])

            _dirty_importers(importers)

            if file_path not in cls.BASE_ADDON_SPECS:
                # Make sure we won't reload importers in a loop if this is actually something
                # that was dynamically imported, where `hot_reload()` might not be called again!
                cls.FILE_MTIMES[file_path] = mtime

    @classmethod
    def hot_reload(cls, mod: Any, require_addons_loaded=False):
        # Solely to trick the type checker because ModuleType doesn't apply where it should
        # and Protocols aren't well-supported yet.
        imported_mod: ModuleType = mod
        imported_file = imported_mod.__file__
        # Mark the caller as having imported (and being dependent on) `module`
        stack = inspect.stack()[1]
        cls.HOTRELOAD_IMPORTERS[imported_file].add(stack.filename)
        cls.FILE_MTIMES[imported_file] = get_mtime(imported_file)

        importing_spec = next((s for s in cls.BASE_ADDON_SPECS if s.origin == stack.filename), None)
        imported_spec = next((s for s in cls.BASE_ADDON_SPECS if s.origin == imported_file), None)

        # Both the importer and the thing being hot-reloaded are registered addons
        if imported_spec and importing_spec:
            # Bad side-effect of our fugly mixed hot-reloading code. If we hot-reload
            # an addon script from another addon, the addon objects for the imported
            # module will never be refreshed.
            if cls.BASE_ADDON_SPECS.index(importing_spec) < cls.BASE_ADDON_SPECS.index(imported_spec):
                raise ImportError(f"{imported_file} addon must be loaded before {stack.filename}"
                                  f" for hot-reloading to work correctly.")

        if require_addons_loaded and not imported_spec:
            # Some addons are dependent not only on the functions from another addon, but additionally
            # require the addons to actually be loaded.
            raise ImportError(f"{stack.filename} expected addons from {imported_file} to be loaded "
                              f"already, but they were not found!")

        # Extra guard against circular imports
        if imported_file in cls._HOT_RELOADING_STACK:
            return

        cls._HOT_RELOADING_STACK.add(imported_file)
        try:
            importlib.reload(imported_mod)
        finally:
            cls._HOT_RELOADING_STACK.remove(imported_file)

    @classmethod
    def get_loaded_script_paths(cls):
        return [s.origin for s in cls.BASE_ADDON_SPECS]

    @classmethod
    def _reload_addons(cls, raise_exceptions=False):
        if cls.LAST_RELOAD and cls.LAST_RELOAD + 2.0 > time.time():
            return
        cls.LAST_RELOAD = time.time()

        cls._check_hotreloads()

        load_exception: Optional[Exception] = None

        new_addons = {}
        for spec in cls.BASE_ADDON_SPECS[:]:
            previous_mod = cls.FRESH_ADDON_MODULES.get(spec.name)
            # Whether we've EVER successfully loaded this module,
            # There may be a `None` entry in the dict if that's the case.
            had_mod = spec.name in cls.FRESH_ADDON_MODULES
            try:
                mtime = get_mtime(spec.origin)
                mtime_changed = mtime != cls.FILE_MTIMES.get(spec.origin, None)
                if not mtime_changed and had_mod:
                    continue
                if not mtime:
                    if not had_mod:
                        # New module and couldn't get mtime? does file exist?
                        raise RuntimeError(f"Couldn't open {spec.origin!r}")
                    else:
                        # Keep module loaded even if file went away.
                        continue

                if previous_mod:
                    cls._unload_module(previous_mod)

                logging.info("(Re)compiling addon %s" % spec.origin)
                mod = importlib.util.module_from_spec(spec)
                sys.modules[spec.name] = mod
                spec.loader.exec_module(mod)
                cls.FILE_MTIMES[spec.origin] = mtime

                new_addons[spec.name] = mod

                # Make sure module initialization happens after any pending task cancellations
                # due to module unloading.
                loop = asyncio.get_event_loop_policy().get_event_loop()
                loop.call_soon(cls._init_module, mod)
            except Exception as e:
                if had_mod:
                    logging.exception("Exploded trying to reload addon %s" % spec.name)
                    cls.FILE_MTIMES.pop(spec.loader, None)
                    cls.FRESH_ADDON_MODULES[spec.name] = None
                else:
                    logging.exception("Exploded trying to load addon %s" % spec.name)
                    cls.BASE_ADDON_SPECS.remove(spec)
                if load_exception is None:
                    load_exception = e
        cls.FRESH_ADDON_MODULES.update(new_addons)
        # if the reload was initialized by a user, let them know that a load failed.
        if raise_exceptions and load_exception is not None:
            raise load_exception

    @classmethod
    def _init_module(cls, mod: ModuleType):
        cls._call_module_hooks(mod, "handle_init", cls.SESSION_MANAGER)
        if not cls._SUBPROCESS:
            for session in cls.SESSION_MANAGER.sessions:
                with addon_ctx.push(new_session=session):
                    cls._call_module_hooks(mod, "handle_session_init", session)

    @classmethod
    def _unload_module(cls, old_mod: ModuleType):
        cls._call_module_hooks(old_mod, "handle_unload", cls.SESSION_MANAGER)
        # Kill any pending tasks that have to die with the old module
        for addon in cls._get_module_addons(old_mod):
            cls.SCHEDULER.kill_matching_tasks(lifetime_mask=TaskLifeScope.ADDON, creator=addon)

    @classmethod
    def _call_all_addon_hooks(cls, hook_name, *args, call_async=False, **kwargs):
        for module in cls.FRESH_ADDON_MODULES.values():
            if not module:
                continue
            ret = cls._call_module_hooks(module, hook_name, *args, call_async=call_async, **kwargs)
            if ret:
                return ret

        return None

    @classmethod
    def _get_module_addons(cls, module):
        return getattr(module, "addons", [])

    @classmethod
    def _get_all_addon_objects(cls):
        addons = []
        for module in cls.FRESH_ADDON_MODULES.values():
            # "module" might really be an addon from the extra_addons param
            addons.append(module)
            addons.extend(cls._get_module_addons(module))
        return addons

    @classmethod
    def _collect_addon_commands(cls) -> Dict[str, WrappedCommandCallable]:
        commands = {}
        for addon in cls._get_all_addon_objects():
            for name, method in inspect.getmembers(addon, predicate=inspect.ismethod):
                command = getattr(method, "command", None)
                if command is None:
                    continue
                commands[name] = method
        return commands

    @classmethod
    def _call_module_hooks(cls, module, hook_name, *args, call_async=False, **kwargs):
        for addon in cls._get_module_addons(module):
            ret = cls._try_call_hook(addon, hook_name, *args, call_async=call_async, **kwargs)
            if ret:
                return ret
        return cls._try_call_hook(module, hook_name, *args, call_async=call_async, **kwargs)

    @classmethod
    def _try_call_hook(cls, addon, hook_name, *args, call_async=False, **kwargs):
        if cls._SUBPROCESS:
            return

        if not addon:
            return
        hook_func = getattr(addon, hook_name, None)
        if not hook_func:
            return
        try:
            if call_async:
                old_hook_func = hook_func

                # Wrapper so we can invoke an async hook synchronously.
                def _wrapper(*w_args, **w_kwargs):
                    cls.SCHEDULER.schedule_task(
                        old_hook_func(*w_args, **w_kwargs),
                        scope=TaskLifeScope.ADDON,
                        creator=addon,
                    )
                    # Fall through to any other handlers as well,
                    # async handlers don't chain.
                    return None
                hook_func = _wrapper
            return hook_func(*args, **kwargs)
        except:
            logging.exception("Exploded in %r's %s hook" % (addon, hook_name))
            if not cls._SWALLOW_ADDON_EXCEPTIONS:
                raise

    @classmethod
    def _show_message(cls, text: str, session: Optional[Session] = None):
        # To deal with circular imports, need to rethink this.
        from hippolyzer.lib.proxy.addon_utils import show_message
        try:
            show_message(text, session=session)
        except:
            if not cls._SWALLOW_ADDON_EXCEPTIONS:
                raise

    @classmethod
    def _show_error(cls, text: str, session: Optional[Session] = None):
        cls._show_message(f"Error: {text}", session=session)
        LOG.error(text)

    @classmethod
    def handle_lludp_message(cls, session: Session, region: ProxiedRegion, message: Message):
        cls._reload_addons()
        if message.name == "ChatFromViewer" and "ChatData" in message:
            if message["ChatData"]["Channel"] == cls.COMMAND_CHANNEL:
                region.circuit.drop_message(message)
                with addon_ctx.push(session, region):
                    try:
                        cls._handle_command(session, region, message["ChatData"]["Message"])
                    except Exception as e:
                        LOG.exception(f'Failed while handling command {message["ChatData"]["Message"]}')
                        cls._show_message(str(e), session)
                        if not cls._SWALLOW_ADDON_EXCEPTIONS:
                            raise
                    return True
        if message.name == "ChatFromSimulator" and "ChatData" in message:
            chat: str = message["ChatData"]["Message"]
            chat_type: int = message["ChatData"]["ChatType"]
            # RLV-style OwnerSay?
            if chat and chat.startswith("@") and chat_type == 8:
                # RLV allows putting multiple commands into one message, blindly splitting on ",".
                chat = chat.lstrip("@")
                all_cmds_handled = True
                for command_str in chat.split(","):
                    if not command_str:
                        continue
                    # RLV-style command, `@<cmd>(:<option1>;<option2>)?(=<param>)?`
                    options, _, param = command_str.partition("=")
                    cmd, _, options = options.partition(":")
                    # TODO: Not always correct, commands can specify their own parsing for the option field
                    options = options.split(";") if options else []
                    source = message["ChatData"]["SourceID"]
                    try:
                        with addon_ctx.push(session, region):
                            handled = cls._call_all_addon_hooks("handle_rlv_command",
                                                                session, region, source, cmd, options, param)
                        if handled:
                            region.circuit.drop_message(message)
                        else:
                            all_cmds_handled = False
                    except:
                        LOG.exception(f"Failed while handling command {command_str!r}")
                        all_cmds_handled = False
                        if not cls._SWALLOW_ADDON_EXCEPTIONS:
                            raise
                # Drop the chat message if all commands it contained were handled by an addon
                if all_cmds_handled:
                    return True

        with addon_ctx.push(session, region):
            return cls._call_all_addon_hooks("handle_lludp_message", session, region, message)

    @classmethod
    def _handle_command(cls, session: Session, region: ProxiedRegion, chat: str):
        commands = cls._collect_addon_commands()
        command, _, param_str = chat.partition(" ")
        if command == "help":
            help_str = "Supported commands:\n"
            for command_name, handler in commands.items():
                command_details: CommandDetails = getattr(handler, 'command')
                stringified_params = ', '.join(command_details.params.keys())
                help_str += f"{command_name}: {stringified_params}\n"
                doc = inspect.getdoc(handler)
                if doc:
                    help_str += textwrap.indent(doc, " " * 4) + "\n"

                # Viewer will truncate around 1600 bytes
                if len(help_str) > 1000:
                    cls._show_message(help_str)
                    help_str = "\n"
            if help_str.strip():
                cls._show_message(help_str)
            return

        handler: Optional[WrappedCommandCallable] = commands.get(command.lower())
        if handler is None:
            cls._show_error(f"No handler for {command!r}")
            return

        command_details: CommandDetails = getattr(handler, 'command')
        addon = getattr(handler, '__self__')
        lifetime = command_details.lifetime
        if lifetime is None:
            lifetime = TaskLifeScope.SESSION | TaskLifeScope.ADDON
        cls.SCHEDULER.schedule_task(
            coro=handler(session, region, param_str),
            scope=lifetime,
            session_id=session.id,
            creator=addon,
        )

    @classmethod
    def handle_http_request(cls, flow: HippoHTTPFlow):
        cls._reload_addons()
        session = flow.cap_data.session() if flow.cap_data.session else None
        region = flow.cap_data.region() if flow.cap_data.region else None
        with addon_ctx.push(session, region):
            return cls._call_all_addon_hooks("handle_http_request", cls.SESSION_MANAGER, flow)

    @classmethod
    def handle_http_response(cls, flow: HippoHTTPFlow):
        cls._reload_addons()
        session = flow.cap_data.session() if flow.cap_data.session else None
        region = flow.cap_data.region() if flow.cap_data.region else None
        with addon_ctx.push(session, region):
            return cls._call_all_addon_hooks("handle_http_response", cls.SESSION_MANAGER, flow)

    @classmethod
    def handle_eq_event(cls, session: Session, region: ProxiedRegion, event: dict):
        cls._reload_addons()
        with addon_ctx.push(session, region):
            return cls._call_all_addon_hooks("handle_eq_event", session, region, event)

    @classmethod
    def handle_session_init(cls, session: Session):
        with addon_ctx.push(session):
            return cls._call_all_addon_hooks("handle_session_init", session)

    @classmethod
    def handle_session_closed(cls, session: Session):
        # kill pending tasks tied to the session
        cls.SCHEDULER.kill_matching_tasks(lifetime_mask=TaskLifeScope.SESSION, session_id=session.id)
        with addon_ctx.push(new_session=session):
            # Specifically give a strong ref so the session doesn't die until all cleanup done
            return cls._call_all_addon_hooks("handle_session_closed", session)

    @classmethod
    def handle_object_updated(cls, session: Session, region: ProxiedRegion,
                              obj: Object, updated_props: Set[str], msg: Optional[Message]):
        with addon_ctx.push(session, region):
            return cls._call_all_addon_hooks("handle_object_updated", session, region, obj, updated_props, msg)

    @classmethod
    def handle_object_killed(cls, session: Session, region: ProxiedRegion, obj: Object):
        with addon_ctx.push(session, region):
            return cls._call_all_addon_hooks("handle_object_killed", session, region, obj)

    @classmethod
    def handle_region_changed(cls, session: Session, region: ProxiedRegion):
        # kill pending tasks tied to the main region
        cls.SCHEDULER.kill_matching_tasks(lifetime_mask=TaskLifeScope.REGION, session_id=session.id)
        with addon_ctx.push(session, region):
            return cls._call_all_addon_hooks("handle_region_changed", session, region)

    @classmethod
    def handle_region_registered(cls, session: Session, region: ProxiedRegion):
        with addon_ctx.push(session, region):
            return cls._call_all_addon_hooks("handle_region_registered", session, region)

    @classmethod
    def handle_circuit_created(cls, session: Session, region: ProxiedRegion):
        with addon_ctx.push(session, region):
            return cls._call_all_addon_hooks("handle_circuit_created", session, region)

    @classmethod
    def handle_proxied_packet(cls, session_manager: SessionManager, packet: UDPPacket,
                              session: Optional[Session], region: Optional[ProxiedRegion]):
        with addon_ctx.push(session, region):
            return cls._call_all_addon_hooks("handle_proxied_packet", session_manager,
                                             packet, session, region)

    @classmethod
    def handle_leap_client_added(cls, session_manager: SessionManager, leap_client: outleap.LEAPClient):
        return cls._call_all_addon_hooks("handle_leap_client_added", session_manager, leap_client, call_async=True)



import abc
import asyncio
import enum
import socket
from typing import *


ADDR_TUPLE = Tuple[str, int]


class Direction(enum.Enum):
    OUT = enum.auto()
    IN = enum.auto()

    def __invert__(self):
        if self == self.OUT:
            return self.IN
        return self.OUT


class UDPPacket:
    def __init__(
            self,
            src_addr: Optional[ADDR_TUPLE],
            dst_addr: ADDR_TUPLE,
            data: bytes,
            direction: Direction
    ):
        self.src_addr = src_addr
        self.dst_addr = dst_addr
        self.data = data
        self.direction = direction
        self.meta = {}

    @property
    def outgoing(self):
        return self.direction == Direction.OUT

    @property
    def incoming(self):
        return self.direction == Direction.IN

    @property
    def far_addr(self):
        if self.outgoing:
            return self.dst_addr
        return self.src_addr

    def __repr__(self):
        return f"<{self.__class__.__name__} src_addr={self.src_addr!r} dst_addr={self.dst_addr!r} data={self.data!r}>"


class AbstractUDPTransport(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    def send_packet(self, packet: UDPPacket) -> None:
        pass

    @abc.abstractmethod
    def close(self) -> None:
        pass


class SocketUDPTransport(AbstractUDPTransport):
    def __init__(self, transport: Union[asyncio.DatagramTransport, socket.socket]):
        super().__init__()
        self.transport = transport

    def send_packet(self, packet: UDPPacket) -> None:
        if not packet.outgoing:
            raise ValueError(f"{self.__class__.__name__} can only send outbound packets")
        self.transport.sendto(packet.data, packet.dst_addr)

    def close(self) -> None:
        self.transport.close()



"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""

import socket
import struct
from typing import *

from hippolyzer.lib.base.datatypes import *
from hippolyzer.lib.base.message.msgtypes import MsgType


PACKER = Callable[[Any], bytes]
UNPACKER = Callable[[bytes], Any]
SPEC = Tuple[UNPACKER, PACKER]


def _pack_string(pack_string):
    """Return the string UTF-8 encoded and null terminated."""
    if pack_string is None:
        return b''
    elif isinstance(pack_string, str):
        return pack_string.encode('utf-8') + b'\x00'
    else:
        return bytes(pack_string)


def _make_struct_spec(struct_fmt: str) -> SPEC:
    struct_obj = struct.Struct(struct_fmt)
    return (lambda x: struct_obj.unpack(x)[0]), struct_obj.pack


def _make_tuplecoord_spec(typ: Type[TupleCoord], struct_fmt: str,
                          needed_elems: Optional[int] = None) -> SPEC:
    struct_obj = struct.Struct(struct_fmt)
    if needed_elems is None:
        # Number of elems needed matches the number in the coord type
        def _packer(x):
            return struct_obj.pack(*x)
    else:
        # Special case, we only want to pack some of the components.
        # Mostly for Quaternion since we don't actually need to send W.
        def _packer(x):
            if isinstance(x, TupleCoord):
                x = x.data()
            return struct_obj.pack(*x[:needed_elems])
    return lambda x: typ(*struct_obj.unpack(x)), _packer


def _unpack_specs(cls):
    cls.UNPACKERS = {k: v[0] for (k, v) in cls.SPECS.items()}
    cls.PACKERS = {k: v[1] for (k, v) in cls.SPECS.items()}
    return cls


@_unpack_specs
class TemplateDataPacker:
    SPECS: Dict[MsgType, SPEC] = {
        MsgType.MVT_FIXED: (bytes, _pack_string),
        MsgType.MVT_VARIABLE: (bytes, _pack_string),
        MsgType.MVT_S8: _make_struct_spec('b'),
        MsgType.MVT_U8: _make_struct_spec('B'),
        MsgType.MVT_BOOL: _make_struct_spec('B'),
        MsgType.MVT_LLUUID: (lambda x: UUID(bytes=bytes(x)), lambda x: UUID(x).bytes),
        MsgType.MVT_IP_ADDR: (socket.inet_ntoa, socket.inet_aton),
        MsgType.MVT_IP_PORT: _make_struct_spec('!H'),
        MsgType.MVT_U16: _make_struct_spec('<H'),
        MsgType.MVT_U32: _make_struct_spec('<I'),
        MsgType.MVT_U64: _make_struct_spec('<Q'),
        MsgType.MVT_S16: _make_struct_spec('<h'),
        MsgType.MVT_S32: _make_struct_spec('<i'),
        MsgType.MVT_S64: _make_struct_spec('<q'),
        MsgType.MVT_F32: _make_struct_spec('<f'),
        MsgType.MVT_F64: _make_struct_spec('<d'),
        MsgType.MVT_LLVector3: _make_tuplecoord_spec(Vector3, "<3f"),
        MsgType.MVT_LLVector3d: _make_tuplecoord_spec(Vector3, "<3d"),
        MsgType.MVT_LLVector4: _make_tuplecoord_spec(Vector4, "<4f"),
        MsgType.MVT_LLQuaternion: _make_tuplecoord_spec(Quaternion, "<3f", needed_elems=3)
    }
    UNPACKERS: Dict[MsgType, UNPACKER] = {}
    PACKERS: Dict[MsgType, PACKER] = {}

    @classmethod
    def unpack(cls, data, data_type):
        return cls.UNPACKERS[data_type](data)

    @classmethod
    def pack(cls, data, data_type):
        return cls.PACKERS[data_type](data)


@_unpack_specs
class LLSDDataPacker(TemplateDataPacker):
    # Some template var types aren't directly representable in LLSD, so they
    # get encoded to binary fields.
    SPECS = {
        MsgType.MVT_IP_ADDR: (socket.inet_ntoa, socket.inet_aton),
        # LLSD ints are technically bound to S32 range.
        MsgType.MVT_U32: _make_struct_spec('!I'),
        MsgType.MVT_U64: _make_struct_spec('!Q'),
        MsgType.MVT_S64: _make_struct_spec('!q'),
    }



import asyncio
import gc
import logging
import multiprocessing
import os
import sys
import time
from typing import Optional

import mitmproxy.ctx
import mitmproxy.exceptions
import outleap

from hippolyzer.lib.base import llsd
from hippolyzer.lib.proxy.addons import AddonManager
from hippolyzer.lib.proxy.addon_utils import BaseAddon
from hippolyzer.lib.proxy.ca_utils import setup_ca
from hippolyzer.lib.proxy.commands import handle_command
from hippolyzer.lib.proxy.http_proxy import create_http_proxy, create_proxy_master, HTTPFlowContext
from hippolyzer.lib.proxy.http_event_manager import MITMProxyEventManager
from hippolyzer.lib.proxy.lludp_proxy import SLSOCKS5Server
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import SessionManager, Session
from hippolyzer.lib.proxy.settings import ProxySettings

LOG = logging.getLogger(__name__)


class SelectionManagerAddon(BaseAddon):
    def handle_lludp_message(self, session: Session, region: ProxiedRegion, message: Message):
        selected = session.selected
        if message.name == "ObjectSelect":
            # ObjectDeselect intentionally ignored to deal with messages that
            # won't work while the object is still selected
            selected.object_local = message["ObjectData"][-1]["ObjectLocalID"]
            needed_objects = set()
            selected.object_locals = tuple(b["ObjectLocalID"] for b in message["ObjectData"])
            for local_id in selected.object_locals:
                obj = region.objects.lookup_localid(local_id)
                if obj:
                    if not obj.ParentID:
                        selected.object_local = obj.LocalID
                else:
                    LOG.debug(f"Don't know about selected {local_id}, requesting object")
                    needed_objects.add(local_id)

            if needed_objects and session.session_manager.settings.ALLOW_AUTO_REQUEST_OBJECTS:
                region.objects.request_objects(needed_objects)
        # ParcelDwellRequests are sent whenever "about land" is opened. This gives us a
        # decent mechanism for selecting parcels.
        elif message.name == "ParcelDwellReply":
            block = message["Data"]
            selected.parcel_full = block["ParcelID"]
            selected.parcel_local = block["LocalID"]
        elif message.name == "UpdateTaskInventory":
            # This isn't quite right because UpdateTaskInventory is basically an
            # upsert, and takes on the insert behaviour if ItemID isn't in the task
            # already. The insert behaviour will give it a _new_ ItemID.
            # We have no way to distinguish the two without additional context.
            # You can always tick the share to group box though, and that should trigger
            # an update with the correct, new item ID.
            selected.task_item = message["InventoryData"]["ItemID"]

    def handle_http_request(self, session_manager, flow):
        cap_data = flow.cap_data
        if not cap_data:
            return
        if not cap_data.session:
            return
        selected = cap_data.session().selected
        # GetMetadata gets called whenever you open a script in a task, so we
        # can use it to guess which script you last opened.
        if cap_data.cap_name == "GetMetadata":
            parsed = llsd.parse_xml(flow.request.content)
            selected.script_item = parsed["item-id"]
            selected.task_item = parsed["item-id"]


class REPLAddon(BaseAddon):
    @handle_command()
    async def spawn_repl(self, session: Session, region: ProxiedRegion):
        """Spawn a REPL"""
        if not AddonManager.have_active_repl():
            AddonManager.spawn_repl()


def run_http_proxy_process(proxy_host, http_proxy_port, flow_context: HTTPFlowContext):
    mitm_loop = asyncio.new_event_loop()
    asyncio.set_event_loop(mitm_loop)

    async def mitmproxy_loop():
        mitmproxy_master = create_http_proxy(proxy_host, http_proxy_port, flow_context)
        gc.freeze()
        await mitmproxy_master.run()

    asyncio.run(mitmproxy_loop())


def start_proxy(session_manager: SessionManager, extra_addons: Optional[list] = None,
                extra_addon_paths: Optional[list] = None, proxy_host=None):
    extra_addons = extra_addons or []
    extra_addon_paths = extra_addon_paths or []
    extra_addons.append(SelectionManagerAddon())
    extra_addons.append(REPLAddon())

    root_log = logging.getLogger()
    root_log.addHandler(logging.StreamHandler())
    root_log.setLevel(logging.INFO)
    logging.basicConfig()

    loop = asyncio.get_event_loop_policy().get_event_loop()

    udp_proxy_port = session_manager.settings.SOCKS_PROXY_PORT
    http_proxy_port = session_manager.settings.HTTP_PROXY_PORT
    leap_port = session_manager.settings.LEAP_PORT
    if proxy_host is None:
        proxy_host = session_manager.settings.PROXY_BIND_ADDR

    flow_context = session_manager.flow_context
    session_manager.name_cache.load_viewer_caches()

    # TODO: argparse
    if len(sys.argv) == 3:
        if sys.argv[1] == "--setup-ca":
            try:
                mitmproxy_master = create_http_proxy(proxy_host, http_proxy_port, flow_context)
            except mitmproxy.exceptions.MitmproxyException:
                # Proxy already running, create the master so we don't try to bind to a port
                mitmproxy_master = create_proxy_master(proxy_host, http_proxy_port, flow_context)
            setup_ca(sys.argv[2], mitmproxy_master)
            return sys.exit(0)

    http_proc = multiprocessing.Process(
        target=run_http_proxy_process,
        args=(proxy_host, http_proxy_port, flow_context),
        daemon=True,
    )
    http_proc.start()
    # These need to be set for mitmproxy's ASGIApp serving code to work.
    mitmproxy.ctx.master = None
    mitmproxy.ctx.log = logging.getLogger("mitmproxy log")

    server = SLSOCKS5Server(session_manager)
    coro = asyncio.start_server(server.handle_connection, proxy_host, udp_proxy_port)
    async_server = loop.run_until_complete(coro)

    leap_server = outleap.LEAPBridgeServer(session_manager.leap_client_connected)
    coro = asyncio.start_server(leap_server.handle_connection, proxy_host, leap_port)
    async_leap_server = loop.run_until_complete(coro)

    event_manager = MITMProxyEventManager(session_manager, flow_context)
    loop.create_task(event_manager.run())

    addon_paths = sys.argv[1:]
    addon_paths.extend(extra_addon_paths)
    AddonManager.init(addon_paths, session_manager, addon_objects=extra_addons)

    # Everything in memory at this point should stay
    gc.freeze()
    gc.set_threshold(5000, 50, 10)

    # Serve requests until Ctrl+C is pressed
    print(f"SOCKS and HTTP proxies running on {proxy_host}")
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    print("Asking mitmproxy to shut down")

    # Shut down mitmproxy
    flow_context.shutdown_signal.set()

    # Close the server
    print("Closing SOCKS server")
    async_server.close()
    print("Shutting down LEAP server")
    async_leap_server.close()
    print("Shutting down addons")
    AddonManager.shutdown()
    print("Waiting for SOCKS server to close")
    loop.run_until_complete(async_server.wait_closed())
    print("Closing event loop")
    if sys.platform == "win32":
        # Hack around some bug with qasync's event loop hanging on win32
        p = multiprocessing.Process(
            target=_windows_timeout_killer,
            args=(os.getpid(),)
        )
        p.daemon = True
        p.start()
    loop.close()


def _windows_timeout_killer(pid: int):
    time.sleep(2.0)
    print("Killing hanging event loop")
    os.kill(pid, 9)


def main():
    multiprocessing.set_start_method("spawn")
    start_proxy(SessionManager(ProxySettings()))


if __name__ == "__main__":
    multiprocessing.freeze_support()
    main()



import copy
from typing import *

from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.message.data_packer import LLSDDataPacker
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.message.template import MessageTemplateVariable
from hippolyzer.lib.base.message.template_dict import TemplateDictionary, DEFAULT_TEMPLATE_DICT

VAR_PAIR = Tuple[dict, MessageTemplateVariable]


class LLSDMessageSerializer:
    DEFAULT_TEMPLATE = DEFAULT_TEMPLATE_DICT

    def __init__(self, message_template=None, message_cls: Type[Message] = Message):
        if message_template is not None:
            self.template_dict = TemplateDictionary(message_template=message_template)
        else:
            self.template_dict = self.DEFAULT_TEMPLATE
        self._message_cls = message_cls

    def _yield_vars(self, msg_dict: dict) -> Generator[VAR_PAIR, None, None]:
        # Things that deserialize to distinct types already have
        # defined serializers in LLSD, but for some types we need
        # to look at the template to figure out how to pack them.
        # See U64, IPADDR and friends.
        template = self.template_dict.get_template_by_name(msg_dict['message'])
        for tmpl_block in template.blocks:
            for block in msg_dict['body'].get(tmpl_block.name, ()):
                for tmpl_var in tmpl_block.variables:
                    if tmpl_var.type in LLSDDataPacker.SPECS:
                        yield block, tmpl_var

    def can_handle(self, msg_name: str):
        return msg_name in self.template_dict

    def serialize(self, msg: Message, as_dict=False) -> dict:
        msg_dict = msg.to_dict()
        for block, tmpl_var in self._yield_vars(msg_dict):
            val = block[tmpl_var.name]
            block[tmpl_var.name] = LLSDDataPacker.pack(val, tmpl_var.type)
        if as_dict:
            return msg_dict
        return llsd.format_xml(msg_dict)

    def deserialize(self, llsd_val: Union[dict, bytes]) -> Message:
        if isinstance(llsd_val, bytes):
            llsd_val = llsd.parse(llsd_val)
        else:
            llsd_val = copy.deepcopy(llsd_val)
        if not isinstance(llsd_val, dict):
            raise ValueError(f"Bad LLSD message: {llsd_val!r}")

        for block, tmpl_var in self._yield_vars(llsd_val):
            val = block[tmpl_var.name]
            block[tmpl_var.name] = LLSDDataPacker.unpack(val, tmpl_var.type)
        return self._message_cls.from_dict(llsd_val)



"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
import struct
import re

from . import template
from .msgtypes import MsgFrequency, MsgTrust, MsgEncoding
from .msgtypes import MsgDeprecation, MsgBlockType, MsgType
from ..exc import MessageTemplateParsingError, MessageTemplateNotFound

MESSAGE = 1
BLOCK = 2
DATA = 3


class MessageTemplateParser:
    """a parser which parses the message template and creates MessageTemplate objects
       which are stored in self.message_templates
    """

    START_RE = re.compile(r'.*{.*')
    END_RE = re.compile(r'.*}.*')
    BLOCK_DATA_RE = re.compile(r'.*?(\w+)\s+(\w+)(\s+(\d+))?.*')
    BLOCK_HEADER_RE = re.compile(r'.*?(\w+)\s+(\w+)(\s+(\d+))?.*')
    MESSAGE_HEADER_RE = re.compile(r'.*?(\w+)\s+(\w+)\s+(\w+)\s+(\w+)\s+(\w+)(\s+(\w+))?.*')
    VERSION_RE = re.compile(r"version.(.+)")
    COMMENT_RE = re.compile(r"^\s*//.*$")

    def __init__(self, template_file):
        if template_file is None:
            raise MessageTemplateNotFound("initializing template parser")

        self.template_file = template_file
        self.message_templates = []
        self.version = ''
        self.count = 0
        self.state = 0
        self._parse_template_file()

    def _add_template(self, new_template):
        self.count += 1
        self.message_templates.append(new_template)

    def _parse_template_file(self):

        # regular expressions
        self.template_file.seek(0)
        lines = self.template_file

        current_template = None
        current_block = None

        for line in lines:
            if self.COMMENT_RE.match(line):
                continue

            start = self.START_RE.match(line)
            end = self.END_RE.match(line)

            if self.version == '':
                version_test = self.VERSION_RE.match(line)  # gets packet headers
                if version_test is not None:
                    parts = version_test.group(1)
                    parts = parts.split()
                    self.version = float(parts[0])

            if start:
                self.state += 1

            if self.state == MESSAGE:
                message_header = self.MESSAGE_HEADER_RE.match(line)
                if message_header is not None:
                    current_template = self._start_new_template(message_header)
                    self._add_template(current_template)

            if self.state == BLOCK:
                block_header = self.BLOCK_HEADER_RE.match(line)
                if block_header is not None:
                    current_block = self._start_new_block(block_header)
                    current_template.add_block(current_block)

            if self.state == DATA:
                block_data = self.BLOCK_DATA_RE.match(line)
                if block_data is not None:
                    current_block.add_variable(self._start_new_var(block_data))

            if end:
                self.state -= 1

    def _start_new_template(self, match):

        new_template = template.MessageTemplate(match.group(1))

        frequency = None
        freq_str = match.group(2)
        if freq_str == 'Low':
            frequency = MsgFrequency.LOW_FREQUENCY_MESSAGE
        elif freq_str == 'Medium':
            frequency = MsgFrequency.MEDIUM_FREQUENCY_MESSAGE
        elif freq_str == 'High':
            frequency = MsgFrequency.HIGH_FREQUENCY_MESSAGE
        elif freq_str == 'Fixed':
            frequency = MsgFrequency.FIXED_FREQUENCY_MESSAGE

        new_template.frequency = frequency

        msg_num = int(match.group(3), 0)
        if frequency == MsgFrequency.FIXED_FREQUENCY_MESSAGE:
            # have to do this because Fixed messages are stored as a long in the template
            msg_num &= 0xff
            msg_num_bytes = struct.pack('!BBBB', 0xff, 0xff, 0xff, msg_num)
        elif frequency == MsgFrequency.LOW_FREQUENCY_MESSAGE:
            msg_num_bytes = struct.pack('!BBH', 0xff, 0xff, msg_num)
        elif frequency == MsgFrequency.MEDIUM_FREQUENCY_MESSAGE:
            msg_num_bytes = struct.pack('!BB', 0xff, msg_num)
        elif frequency == MsgFrequency.HIGH_FREQUENCY_MESSAGE:
            msg_num_bytes = struct.pack('!B', msg_num)
        else:
            raise Exception("don't know about frequency %s" % frequency)

        new_template.msg_num = msg_num
        new_template.msg_freq_num_bytes = msg_num_bytes

        msg_trust = None
        msg_trust_str = match.group(4)
        if msg_trust_str == 'Trusted':
            msg_trust = MsgTrust.LL_TRUSTED
        elif msg_trust_str == 'NotTrusted':
            msg_trust = MsgTrust.LL_NOTRUST

        new_template.msg_trust = msg_trust

        msg_encoding = None
        msg_encoding_str = match.group(5)
        if msg_encoding_str == 'Unencoded':
            msg_encoding = MsgEncoding.LL_UNENCODED
        elif msg_encoding_str == 'Zerocoded':
            msg_encoding = MsgEncoding.LL_ZEROCODED

        new_template.msg_encoding = msg_encoding

        msg_dep = None
        msg_dep_str = match.group(7)
        if msg_dep_str:
            if msg_dep_str == 'Deprecated':
                msg_dep = MsgDeprecation.LL_DEPRECATED
            elif msg_dep_str == 'UDPDeprecated':
                msg_dep = MsgDeprecation.LL_UDPDEPRECATED
            elif msg_dep_str == 'UDPBlackListed':
                msg_dep = MsgDeprecation.LL_UDPBLACKLISTED
            elif msg_dep_str == 'NotDeprecated':
                msg_dep = MsgDeprecation.LL_NOTDEPRECATED
        else:
            msg_dep = MsgDeprecation.LL_NOTDEPRECATED
        if msg_dep is None:
            raise MessageTemplateParsingError("Unknown msg_dep field %s" % match.group(0))
        new_template.msg_deprecation = msg_dep

        return new_template

    def _start_new_block(self, match):

        new_block = template.MessageTemplateBlock(match.group(1))

        block_type = None
        block_num = 0

        block_type_str = match.group(2)
        if block_type_str == 'Single':
            block_type = MsgBlockType.MBT_SINGLE
        elif block_type_str == 'Multiple':
            block_type = MsgBlockType.MBT_MULTIPLE
            block_num = int(match.group(4))
        elif block_type_str == 'Variable':
            block_type = MsgBlockType.MBT_VARIABLE

        new_block.block_type = block_type
        new_block.number = block_num

        return new_block

    def _start_new_var(self, match):

        type_string = match.group(2)
        var_type = None
        var_size = -1
        if type_string == 'U8':
            var_type = MsgType.MVT_U8
        elif type_string == 'U16':
            var_type = MsgType.MVT_U16
        elif type_string == 'U32':
            var_type = MsgType.MVT_U32
        elif type_string == 'U64':
            var_type = MsgType.MVT_U64
        elif type_string == 'S8':
            var_type = MsgType.MVT_S8
        elif type_string == 'S16':
            var_type = MsgType.MVT_S16
        elif type_string == 'S32':
            var_type = MsgType.MVT_S32
        elif type_string == 'S64':
            var_type = MsgType.MVT_S64
        elif type_string == 'F32':
            var_type = MsgType.MVT_F32
        elif type_string == 'F64':
            var_type = MsgType.MVT_F64
        elif type_string == 'LLVector3':
            var_type = MsgType.MVT_LLVector3
        elif type_string == 'LLVector3d':
            var_type = MsgType.MVT_LLVector3d
        elif type_string == 'LLVector4':
            var_type = MsgType.MVT_LLVector4
        elif type_string == 'LLQuaternion':
            var_type = MsgType.MVT_LLQuaternion
        elif type_string == 'LLUUID':
            var_type = MsgType.MVT_LLUUID
        elif type_string == 'BOOL':
            var_type = MsgType.MVT_BOOL
        elif type_string == 'IPADDR':
            var_type = MsgType.MVT_IP_ADDR
        elif type_string == 'IPPORT':
            var_type = MsgType.MVT_IP_PORT
        elif type_string == 'Fixed' or type_string == 'Variable':
            if type_string == 'Fixed':
                var_type = MsgType.MVT_FIXED
            elif type_string == 'Variable':
                var_type = MsgType.MVT_VARIABLE

            var_size = int(match.group(4))
            if var_size <= 0:
                raise MessageTemplateParsingError("variable size %s does not match %s" % (var_size, type_string))
        # if the size hasn't been read yet, then read it from message_types
        if var_size == -1:
            var_size = var_type.size

        return template.MessageTemplateVariable(match.group(1),
                                                var_type, var_size)



import os
import tempfile
from io import BytesIO

import defusedxml.ElementTree
from glymur import jp2box, Jp2k

# Replace glymur's ElementTree with a safe one
jp2box.ET = defusedxml.ElementTree


class BufferedJp2k(Jp2k):
    """
    For manipulating JP2K from within a binary buffer.

    For many operations glymur expects to be able to re-read from a file
    based on filename, so this is the least brittle approach.
    """

    def __init__(self, contents: bytes):
        stream = BytesIO(contents)
        self.temp_file = tempfile.NamedTemporaryFile(delete=False)
        stream.seek(0)
        self.temp_file.write(stream.read())
        # Windows NT can't handle two FHs open on a tempfile at once
        self.temp_file.close()

        super().__init__(self.temp_file.name)

    def __del__(self):
        if self.temp_file is not None:
            os.remove(self.temp_file.name)
        self.temp_file = None

    def _populate_cparams(self, img_array):
        if self._cratios is None:
            self._cratios = (1920.0, 480.0, 120.0, 30.0, 10.0)
        if self._irreversible is None:
            self.irreversible = True
        return super()._populate_cparams(img_array)

    def __bytes__(self):
        with open(self.temp_file.name, "rb") as f:
            return f.read()



"""
http://wiki.secondlife.com/wiki/Mesh/Mesh_Asset_Format
"""

from __future__ import annotations

import dataclasses
import datetime as dt
import logging
from typing import *
import zlib
from copy import deepcopy

import numpy as np
import recordclass

from hippolyzer.lib.base import serialization as se
from hippolyzer.lib.base.datatypes import Vector3, Vector2, UUID, TupleCoord
from hippolyzer.lib.base.llsd import zip_llsd, unzip_llsd
from hippolyzer.lib.base.serialization import ParseContext

LOG = logging.getLogger(__name__)


def llsd_to_mat4(mat: Union[np.ndarray, Sequence[float]]) -> np.ndarray:
    return np.array(mat).reshape((4, 4), order='F')


def mat4_to_llsd(mat: np.ndarray) -> List[float]:
    return list(mat.flatten(order='F'))


@dataclasses.dataclass
class MeshAsset:
    header: MeshHeaderDict = dataclasses.field(default_factory=dict)
    segments: MeshSegmentDict = dataclasses.field(default_factory=dict)
    raw_segments: Dict[str, bytes] = dataclasses.field(default_factory=dict)

    @classmethod
    def make_triangle(cls) -> MeshAsset:
        """Make an asset representing an un-rigged single-sided mesh triangle"""
        inst = cls()
        inst.header = {
            "version": 1,
            "high_lod": {"offset": 0, "size": 0},
            "physics_mesh": {"offset": 0, "size": 0},
            "physics_convex": {"offset": 0, "size": 0},
        }
        base_lod: LODSegmentDict = {
            'Normal': [
                Vector3(-0.0, -0.0, -1.0),
                Vector3(-0.0, -0.0, -1.0),
                Vector3(-0.0, -0.0, -1.0)
            ],
            'PositionDomain': {'Max': [0.5, 0.5, 0.0], 'Min': [-0.5, -0.5, 0.0]},
            'Position': [
                Vector3(0.0, 0.0, 0.0),
                Vector3(1.0, 0.0, 0.0),
                Vector3(0.5, 1.0, 0.0)
            ],
            'TexCoord0Domain': {'Max': [1.0, 1.0], 'Min': [0.0, 0.0]},
            'TexCoord0': [
                Vector2(0.0, 0.0),
                Vector2(1.0, 0.0),
                Vector2(0.5, 1.0)
            ],
            'TriangleList': [[0, 1, 2]],
        }
        inst.segments['physics_mesh'] = [deepcopy(base_lod)]
        inst.segments['high_lod'] = [deepcopy(base_lod)]
        convex_segment: PhysicsConvexSegmentDict = {
            'BoundingVerts': [
                Vector3(-0.0, 1.0, -1.0),
                Vector3(-1.0, -1.0, -1.0),
                Vector3(1.0, -1.0, -1.0)
            ],
            'Max': [0.5, 0.5, 0.0],
            'Min': [-0.5, -0.5, 0.0]
        }
        inst.segments['physics_convex'] = convex_segment
        return inst

    def iter_lods(self) -> Generator[List[LODSegmentDict], None, None]:
        for lod_name, lod_val in self.segments.items():
            if lod_name.endswith("_lod"):
                yield lod_val

    def iter_lod_materials(self) -> Generator[LODSegmentDict, None, None]:
        for lods in self.iter_lods():
            yield from lods


# These TypedDicts describe the expected shape of the LLSD in the mesh
# header and various segments. They're mainly for type hinting.
class MeshHeaderDict(TypedDict, total=False):
    """Header of the mesh file, includes offsets & sizes for segments' LLSD"""
    version: int
    creator: UUID
    date: dt.datetime
    physics_cost_data: PhysicsCostDataHeaderDict
    high_lod: LODSegmentHeaderDict
    medium_lod: LODSegmentHeaderDict
    low_lod: LODSegmentHeaderDict
    lowest_lod: LODSegmentHeaderDict
    physics_convex: PhysicsSegmentHeaderDict
    physics_mesh: PhysicsSegmentHeaderDict
    skin: SegmentHeaderDict
    physics_havok: PhysicsHavokSegmentHeaderDict


class SegmentHeaderDict(TypedDict):
    """Standard shape for segment references within the header"""
    offset: int
    size: int


class LODSegmentHeaderDict(SegmentHeaderDict, total=False):
    # Possibly only on lowest?
    mesh_triangles: int


class PhysicsSegmentHeaderDict(SegmentHeaderDict, total=False):
    # Populated by the server, not there on the client
    hash: bytes


class PhysicsHavokSegmentHeaderDict(PhysicsSegmentHeaderDict, total=False):
    version: int


class PhysicsCostDataHeaderDict(TypedDict, total=False):
    """Cost of physical representation, populated by server"""
    decomposition: float
    decomposition_discounted_vertices: int
    decomposition_hulls: int
    hull: float
    hull_discounted_vertices: float
    # Not sure what this is, always seems to be 9 floats -1 to 1 if present
    # Mat3?
    mesh: List[float]
    mesh_triangles: int


class MeshSegmentDict(TypedDict, total=False):
    """Dict of segments unpacked using the MeshHeaderDict"""
    high_lod: List[LODSegmentDict]
    medium_lod: List[LODSegmentDict]
    low_lod: List[LODSegmentDict]
    lowest_lod: List[LODSegmentDict]
    physics_convex: PhysicsConvexSegmentDict
    physics_mesh: List[LODSegmentDict]
    physics_havok: PhysicsHavokSegmentDict
    skin: SkinSegmentDict


class LODSegmentDict(TypedDict, total=False):
    """Represents a single entry within the material list of a LOD segment"""
    # Only present if True and no geometry
    NoGeometry: bool
    # -1.0 - 1.0
    Position: List[Vector3]
    PositionDomain: DomainDict
    # 0.0 - 1.0
    TexCoord0: List[Vector2]
    TexCoord0Domain: DomainDict
    # -1.0 - 1.0
    Normal: List[Vector3]
    # [[1,2,3], [1,3,4], ...]
    TriangleList: List[List[int]]
    # Only present if rigged
    Weights: List[List[VertexWeight]]


class DomainDict(TypedDict):
    """Description of the real range for quantized coordinates"""
    # number of elems depends on what the domain is for, Vec2 or Vec3
    Max: List[float]
    Min: List[float]


class VertexWeight(recordclass.RecordClass):
    """Vertex weight for a specific joint on a specific vertex"""
    # index of the joint within the joint_names list in the skin segment
    joint_idx: int
    # 0.0 - 1.0
    weight: float


class SkinSegmentDict(TypedDict, total=False):
    """Rigging information"""
    joint_names: List[str]
    # model -> world transform mat4 for model
    bind_shape_matrix: List[float]
    # world -> joint local transform mat4s
    inverse_bind_matrix: List[List[float]]
    # Transform mat4s for the joint nodes themselves.
    # The matrices may have scale or other components, but only the
    # translation component will be used by the viewer.
    # All translations are relative to the joint's parent.
    alt_inverse_bind_matrix: List[List[float]]
    lock_scale_if_joint_position: bool
    pelvis_offset: float


class PhysicsConvexSegmentDict(DomainDict, total=False):
    """
    Data for convex hull collisions, populated by the client

    Min / Max pos domain vals are inline, unlike for LODs, so this inherits from DomainDict
    """
    # Indices into the Positions list
    HullList: List[int]
    # -1.0 - 1.0, dequantized from binary field of U16s
    Positions: List[Vector3]
    # -1.0 - 1.0, dequantized from binary field of U16s
    BoundingVerts: List[Vector3]


class PhysicsHavokSegmentDict(TypedDict, total=False):
    """Cached data for Havok collisions, populated by sim and not used by client."""
    HullMassProps: HavokMassPropsDict
    MOPP: HavokMOPPDict
    MeshDecompMassProps: HavokMassPropsDict
    WeldingData: bytes


class HavokMassPropsDict(TypedDict, total=False):
    # Vec, center of mass
    CoM: List[float]
    # 9 floats, Mat3?
    inertia: List[float]
    mass: float
    volume: float


class HavokMOPPDict(TypedDict, total=False):
    """Memory Optimized Partial Polytope"""
    BuildType: int
    MoppData: bytes
    # 4 floats, Vec4?
    MoppInfo: List[float]


def positions_from_domain(positions: Iterable[TupleCoord], domain: DomainDict):
    """
    Used for turning positions into their actual positions within the mesh / domain

    for ex: positions_from_domain(lod["Position"], lod["PositionDomain])
    """
    lower = domain['Min']
    upper = domain['Max']
    return [
        x.interpolate(lower, upper) for x in positions
    ]


def positions_to_domain(positions: Iterable[TupleCoord], domain: DomainDict):
    """Used for turning positions into their actual positions within the mesh / domain"""
    lower = domain['Min']
    upper = domain['Max']
    return [
        x.within_domain(lower, upper) for x in positions
    ]


class VertexWeights(se.SerializableBase):
    """Serializer for a list of joint weights on a single vertex"""
    INFLUENCE_LIMIT = 4
    INFLUENCE_TERM = 0xFF

    @classmethod
    def serialize(cls, vals, writer: se.BufferWriter, ctx=None):
        if len(vals) > cls.INFLUENCE_LIMIT:
            raise ValueError(f"{vals!r} is too long, can only have {cls.INFLUENCE_LIMIT} influences!")
        for val in vals:
            joint_idx, influence = val
            writer.write(se.U8, joint_idx)
            writer.write(se.U16, round(influence * 0xFFff), ctx=ctx)
        if len(vals) != cls.INFLUENCE_LIMIT:
            writer.write(se.U8, cls.INFLUENCE_TERM)

    @classmethod
    def deserialize(cls, reader: se.Reader, ctx=None):
        # NOTE: normally you'd want to do something like arrange this into a nicely
        # aligned byte array with zero padding so that you could vectorize the decoding.
        # In cases where having a vertex with no weights is semantically equivalent to
        # having a vertex _with_ weights of a value of 0.0 that's fine. This isn't the case
        # in LL's implementation of mesh:
        #
        # https://bitbucket.org/lindenlab/viewer/src/d31a83fb946c49a38376ea3b312b5380d0c8c065/indra/llmath/llvolume.cpp#lines-2560:2628
        #
        # Consider the difference between handling of b"\x00\x00\x00\xFF" and b"\xFF" with the above logic.
        # To simplify round-tripping while preserving those semantics, we don't do a vectorized decode.
        # I had a vectorized numpy version, but those requirements made everything a bit of a mess.
        influence_list = []
        for _ in range(cls.INFLUENCE_LIMIT):
            joint_idx = reader.read_bytes(1)[0]
            if joint_idx == cls.INFLUENCE_TERM:
                break
            weight = reader.read(se.U16, ctx=ctx) / 0xFFff
            influence_list.append(VertexWeight(joint_idx, weight))
        return influence_list


class SegmentSerializer:
    """Serializer for binary fields within an LLSD object"""
    def __init__(self, templates):
        self._templates: Dict[str, se.SerializableBase] = templates

    def serialize(self, vals: Dict[str, Any]):
        new_segment = {}
        for key, val in vals.items():
            if key in self._templates and not isinstance(val, bytes):
                # Pretty much everything in mesh segments is little-endian other
                # than the LLSD itself.
                writer = se.BufferWriter("<")
                writer.write(self._templates[key], val)
                new_segment[key] = writer.copy_buffer()
            else:
                new_segment[key] = val
        return new_segment

    def deserialize(self, vals: Dict[str, Any]):
        new_segment = {}
        for key, val in vals.items():
            if key in self._templates:
                reader = se.BufferReader("<", val)
                new_segment[key] = reader.read(self._templates[key])
                if len(reader):
                    LOG.warning(f"{len(reader)} bytes left in reader on mesh key {key}")
            else:
                new_segment[key] = val
        return new_segment


class VecListAdapter(se.Adapter):
    def __init__(self, child_spec: se.SERIALIZABLE_TYPE, vec_type: Type):
        super().__init__(child_spec)
        self.vec_type = vec_type

    def encode(self, val: Any, ctx: Optional[ParseContext]) -> Any:
        return val

    def decode(self, val: Any, ctx: Optional[ParseContext], pod: bool = False) -> Any:
        new_vals = []
        for elem in val:
            new_vals.append(self.vec_type(*elem))
        return new_vals


LE_U16: np.dtype = np.dtype(np.uint16).newbyteorder('<')  # noqa


LOD_SEGMENT_SERIALIZER = SegmentSerializer({
    # 16-bit indices to the verts making up the tri. Imposes a 16-bit
    # upper limit on verts in any given material in the mesh.
    "TriangleList": se.ExprAdapter(
        se.NumPyArray(se.BytesGreedy(), LE_U16, 3),
        decode_func=lambda x: x.tolist(),
    ),
    # These are used to interpolate between values in their respective domains
    # Each position represents a single vert.
    "Position": VecListAdapter(
        se.QuantizedNumPyArray(se.NumPyArray(se.BytesGreedy(), LE_U16, 3), 0.0, 1.0),
        Vector3,
    ),
    "TexCoord0": VecListAdapter(
        se.QuantizedNumPyArray(se.NumPyArray(se.BytesGreedy(), LE_U16, 2), 0.0, 1.0),
        Vector2,
    ),
    # Normals have a static domain between -1 and 1, so we just use that rather than 0.0 - 1.0.
    "Normal": VecListAdapter(
        se.QuantizedNumPyArray(se.NumPyArray(se.BytesGreedy(), LE_U16, 3), -1.0, 1.0),
        Vector3,
    ),
    "Weights": se.Collection(None, VertexWeights)
})


class LLMeshSerializer(se.SerializableBase):
    # Also used as serialization order for segments.
    # Note that there's conflicting info about whether skin is supposed to
    # come first or not but the viewer always puts it second last, so we will too.
    KNOWN_SEGMENTS = ("lowest_lod", "low_lod", "medium_lod", "high_lod",
                      "physics_mesh", "physics_convex", "skin", "physics_havok")

    # Define unpackers for specific binary fields within the parsed LLSD segments
    SEGMENT_TEMPLATES: Dict[str, SegmentSerializer] = {
        "lowest_lod": LOD_SEGMENT_SERIALIZER,
        "low_lod": LOD_SEGMENT_SERIALIZER,
        "medium_lod": LOD_SEGMENT_SERIALIZER,
        "high_lod": LOD_SEGMENT_SERIALIZER,
        "physics_mesh": LOD_SEGMENT_SERIALIZER,
        "physics_convex": SegmentSerializer({
            "BoundingVerts": se.Collection(None, se.Vector3U16(-1.0, 1.0)),
            "HullList": se.Collection(None, se.U8),
            "Positions": se.Collection(None, se.Vector3U16(-1.0, 1.0)),
        }),
    }

    def __init__(
        self,
        parse_segment_contents: bool = True,
        allow_invalid_segments: bool = False,
        include_raw_segments: bool = False,
    ):
        super().__init__()
        self.parse_segment_contents = parse_segment_contents
        self.allow_invalid_segments = allow_invalid_segments
        self.include_raw_segments = include_raw_segments

    @classmethod
    def _segment_sort(cls, key):
        if key in cls.KNOWN_SEGMENTS:
            return cls.KNOWN_SEGMENTS.index(key)
        # If we don't know what this is chuck it to the end
        return 0xFFffFFff

    def _is_segment_header(self, val):
        if not isinstance(val, dict):
            return False
        if "offset" not in val:
            return False
        if "size" not in val:
            return False
        return True

    def serialize(self, val: MeshAsset, writer: se.BufferWriter, ctx=None):
        all_segs = set(val.segments.keys()) | set(val.raw_segments.keys())
        missing_headers = all_segs - set(val.header.keys())
        if missing_headers:
            raise ValueError(f"Segments missing from header dict: {missing_headers}")

        # We write the body first so we can get the offsets and sizes
        # for each segment for the header.
        inner_writer = se.BufferWriter(writer.endianness)
        # Don't mutate the original header
        new_header = deepcopy(val.header)

        # Write the segments in their preferred order
        for key in sorted(new_header.keys(), key=self._segment_sort):
            segment_header = new_header[key]  # type: ignore
            if not self._is_segment_header(segment_header):
                # Doesn't look like a segment header
                continue
            if key not in self.KNOWN_SEGMENTS:
                # Serialize anyway, it's all LLSD.
                LOG.warning(f"Serializing unknown mesh segment {key}")

            # Try segments first, then raw_segments
            segment_val = val.segments.get(key, val.raw_segments.get(key))  # type: ignore
            if segment_val is None:
                if self.allow_invalid_segments:
                    continue
                raise ValueError(f"{key} segment in header missing from segments dict")
            start_offset = len(inner_writer)

            # Write the segment, updating the header with the new offsets and sizes
            segment_header["offset"] = start_offset
            if isinstance(segment_val, bytes):
                inner_writer.write_bytes(segment_val)
            else:
                if key in self.SEGMENT_TEMPLATES:
                    if isinstance(segment_val, (list, tuple)):
                        segment_val = [self.SEGMENT_TEMPLATES[key].serialize(x) for x in segment_val]
                    else:
                        segment_val = self.SEGMENT_TEMPLATES[key].serialize(segment_val)  # type: ignore
                inner_writer.write_bytes(zip_llsd(segment_val))
            segment_header["size"] = len(inner_writer) - start_offset
        writer.write(se.BinaryLLSD, new_header, ctx=ctx)
        writer.write_bytes(inner_writer.buffer)

    def deserialize(self, reader: se.Reader, ctx=None):
        mesh = MeshAsset()
        mesh.header = reader.read(se.BinaryLLSD, ctx=ctx)
        header_end = reader.tell()
        bytes_after_header = len(reader)
        for key, segment_header in mesh.header.items():
            if not self._is_segment_header(segment_header):
                # Doesn't look like a segment header
                continue
            if key not in self.KNOWN_SEGMENTS:
                LOG.warning(f"Encountered unknown mesh segment on decode: {key}")
            if segment_header['offset'] + segment_header['size'] > bytes_after_header:
                err_msg = f"{segment_header!r} would pass EOF, refusing to parse"
                if self.allow_invalid_segments:
                    LOG.debug(err_msg)
                    continue
                else:
                    raise ValueError(err_msg)
            reader.seek(header_end + segment_header['offset'])
            seg_bytes = reader.read_bytes(segment_header['size'])

            if self.allow_invalid_segments and all(b == 0x00 for b in seg_bytes):
                LOG.debug("Encountered padding segment, skipping")
                continue
            try:
                segment_llsd = unzip_llsd(seg_bytes)
            except zlib.error:
                if self.allow_invalid_segments:
                    LOG.debug(f"Failed to parse segment bytes for {key}")
                    continue
                raise
            if self.parse_segment_contents and key in self.SEGMENT_TEMPLATES:
                if isinstance(segment_llsd, (list, tuple)):
                    segment_parsed = [self.SEGMENT_TEMPLATES[key].deserialize(x) for x in segment_llsd]
                else:
                    segment_parsed = self.SEGMENT_TEMPLATES[key].deserialize(segment_llsd)
            else:
                segment_parsed = segment_llsd
            mesh.segments[key] = segment_parsed  # type: ignore
            if self.include_raw_segments:
                mesh.raw_segments[key] = seg_bytes
        return mesh



import asyncio
import enum
import weakref
from typing import *

from hippolyzer.lib.base.datatypes import UUID


class TaskLifeScope(enum.Flag):
    """Task should be automatically canceled when data related to flag is changed"""
    # Cancel task when session is closed
    SESSION = enum.auto()
    # Cancel task when _main_ region changes
    REGION = enum.auto()
    # Cancel task when the object that created it (usually an addon) is unloaded
    # (all tasks are canceled when proxy is closed regardless)
    ADDON = enum.auto()


class TaskLifeData:
    def __init__(
            self,
            scope: TaskLifeScope,
            session_id: Optional[UUID] = None,
            creator: Optional[Any] = None,
    ):
        if scope & (TaskLifeScope.REGION | TaskLifeScope.SESSION) and not session_id:
            raise ValueError(f"{scope!r} requires non-null session_id")
        elif scope & TaskLifeScope.ADDON and not creator:
            raise ValueError(f"{scope!r} requires non-null creator addon object")

        # Region-scoped implies session-scoped
        if scope & TaskLifeScope.REGION:
            scope |= TaskLifeScope.SESSION
        self.scope = scope
        self.session_id = session_id
        # only needed for looking for tasks created by this object
        self.creator = weakref.proxy(creator) if creator else None


class TaskScheduler:
    def __init__(self):
        self.tasks: List[Tuple[TaskLifeData, asyncio.Task]] = []

    @staticmethod
    async def _ignore_coro_cancellation(coro: Coroutine):
        try:
            await coro
        except asyncio.CancelledError:
            # If the task didn't handle its own CancelledError
            # then we don't care.
            pass

    def schedule_task(self, coro: Coroutine, scope: Optional[TaskLifeScope] = None,
                      session_id: Optional[UUID] = None, creator: Any = None):
        scope = scope or TaskLifeScope(0)
        task_data = TaskLifeData(scope, session_id, creator)
        task = asyncio.create_task(self._ignore_coro_cancellation(coro))
        task.add_done_callback(self._task_done)
        self.tasks.append((task_data, task))
        return task

    def shutdown(self):
        for task_data, task in self.tasks:
            task.cancel()

        try:
            event_loop = asyncio.get_running_loop()
            await_all = asyncio.gather(*(task for task_data, task in self.tasks))
            event_loop.run_until_complete(await_all)
        except RuntimeError:
            pass
        self.tasks.clear()

    def _task_done(self, task: asyncio.Task):
        for task_details in reversed(self.tasks):
            if task == task_details[1]:
                self.tasks.remove(task_details)
                break

    def get_matching_tasks(self, creator=None, session_id=None):
        for task_data, task in self.tasks[:]:
            if creator and creator == task_data.creator:
                yield task_data, task
            elif session_id and session_id == task_data.session_id:
                yield task_data, task

    def kill_matching_tasks(self, lifetime_mask: TaskLifeScope, **kwargs):
        for task_data, task in self.get_matching_tasks(**kwargs):
            if task_data.scope & lifetime_mask:
                task.cancel()



"""
Base classes for common session-related state shared between clients and proxies
"""
from __future__ import annotations

import abc
from typing import *

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.circuit import ConnectionHolder
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.message.message_handler import MessageHandler
from hippolyzer.lib.base.network.caps_client import CapsClient
from hippolyzer.lib.base.network.transport import ADDR_TUPLE

if TYPE_CHECKING:
    from hippolyzer.lib.client.object_manager import ClientObjectManager, ClientWorldObjectManager


class BaseClientRegion(ConnectionHolder, abc.ABC):
    """Represents a client's view of a remote region"""
    handle: Optional[int]
    # Actually a weakref
    session: Callable[[], BaseClientSession]
    objects: ClientObjectManager
    caps_client: CapsClient


class BaseClientSession(abc.ABC):
    """Represents a client's view of a remote session"""
    id: UUID
    agent_id: UUID
    secure_session_id: UUID
    message_handler: MessageHandler[Message, str]
    regions: Sequence[BaseClientRegion]
    region_by_handle: Callable[[int], Optional[BaseClientRegion]]
    region_by_circuit_addr: Callable[[ADDR_TUPLE], Optional[BaseClientRegion]]
    objects: ClientWorldObjectManager
    login_data: Dict[str, Any]



"""
Assorted utilities to make creating animations from scratch easier
"""

import copy
from typing import List, Union

from hippolyzer.lib.base.datatypes import Vector3, Quaternion
from hippolyzer.lib.base.llanim import PosKeyframe, RotKeyframe


def smooth_step(t: float):
    t = max(0.0, min(1.0, t))
    return t * t * (3 - 2 * t)


def rot_interp(r0: Quaternion, r1: Quaternion, t: float):
    """
    Bad quaternion interpolation

    TODO: This is definitely not correct yet seems to work ok? Implement slerp.
    """
    # Ignore W
    r0 = r0.data(3)
    r1 = r1.data(3)
    return Quaternion(*map(lambda pair: ((pair[0] * (1.0 - t)) + (pair[1] * t)), zip(r0, r1)))


def unique_frames(frames: List[Union[PosKeyframe, RotKeyframe]]):
    """Drop frames where time and coordinate are exact duplicates of another frame"""
    new_frames = []
    for frame in frames:
        # TODO: fudge factor for float comparison instead
        if frame not in new_frames:
            new_frames.append(frame)
    return new_frames


def shift_keyframes(frames: List[Union[PosKeyframe, RotKeyframe]], num: int):
    """
    Shift keyframes around by `num` frames

    Assumes keyframes occur at a set cadence, and that first and last keyframe are at the same coord.
    """

    # Get rid of duplicate frames
    frames = unique_frames(frames)
    pop_idx = -1
    insert_idx = 0
    if num < 0:
        insert_idx = len(frames) - 1
        pop_idx = 0
        num = -num
    old_times = [f.time for f in frames]
    new_frames = frames.copy()
    # Drop last, duped frame. We'll copy the first frame to replace it later
    new_frames.pop(-1)
    for _ in range(num):
        new_frames.insert(insert_idx, new_frames.pop(pop_idx))

    # Put first frame back on the end
    new_frames.append(copy.copy(new_frames[0]))

    assert len(old_times) == len(new_frames)
    assert new_frames[0] == new_frames[-1]
    # Make the times of the shifted keyframes match up with the previous timeline
    for old_time, new_frame in zip(old_times, new_frames):
        new_frame.time = old_time
    return new_frames


def smooth_pos(start: Vector3, end: Vector3, inter_frames: int, time: float, duration: float) -> List[PosKeyframe]:
    """Generate keyframes to smoothly interpolate between two positions"""
    frames = [PosKeyframe(time=time, pos=start)]
    for i in range(0, inter_frames):
        t = (i + 1) / (inter_frames + 1)
        smooth_t = smooth_step(t)
        pos = Vector3(smooth_t, smooth_t, smooth_t).interpolate(start, end)
        frames.append(PosKeyframe(time=time + (t * duration), pos=pos))
    return frames + [PosKeyframe(time=time + duration, pos=end)]


def smooth_rot(start: Quaternion, end: Quaternion, inter_frames: int, time: float, duration: float)\
        -> List[RotKeyframe]:
    """Generate keyframes to smoothly interpolate between two rotations"""
    frames = [RotKeyframe(time=time, rot=start)]
    for i in range(0, inter_frames):
        t = (i + 1) / (inter_frames + 1)
        smooth_t = smooth_step(t)
        frames.append(RotKeyframe(time=time + (t * duration), rot=rot_interp(start, end, smooth_t)))
    return frames + [RotKeyframe(time=time + duration, rot=end)]



# This currently implements basic LLMesh -> Collada.
#
# TODO:
#  * inverse, Collada -> LLMesh (for simple cases, maybe using impasse rather than pycollada)
#  * round-tripping tests, LLMesh->Collada->LLMesh
#  * * Can't really test using Collada->LLMesh->Collada because Collada->LLMesh is almost always
#      going to be lossy due to how SL represents vertex data and materials compared to what
#      Collada allows.
#  * Eventually scrap this and just use GLTF instead once we know we have the semantics correct
#  * * Collada was just easier to bootstrap given that it's the only officially supported input format
#  * * Collada tooling sucks and even LL is moving away from it
#  * * Ensuring LLMesh->Collada and LLMesh->GLTF conversion don't differ semantically is easy via assimp.

import logging
import os.path
import secrets
import sys
from typing import Dict, Optional

import collada
import collada.source
from collada import E
from lxml import etree
import numpy as np
import transformations

from hippolyzer.lib.base.helpers import get_resource_filename
from hippolyzer.lib.base.serialization import BufferReader
from hippolyzer.lib.base.mesh import (
    LLMeshSerializer,
    MeshAsset,
    positions_from_domain,
    SkinSegmentDict,
    llsd_to_mat4,
)

LOG = logging.getLogger(__name__)
DIR = os.path.dirname(os.path.realpath(__file__))


def mat4_to_collada(mat: np.ndarray) -> np.ndarray:
    return mat.flatten(order='C')


def mesh_to_collada(ll_mesh: MeshAsset, include_skin=True) -> collada.Collada:
    dae = collada.Collada()
    axis = collada.asset.UP_AXIS.Z_UP
    dae.assetInfo.upaxis = axis
    scene = collada.scene.Scene("scene", [llmesh_to_node(ll_mesh, dae, include_skin=include_skin)])

    dae.scenes.append(scene)
    dae.scene = scene
    return dae


def llmesh_to_node(ll_mesh: MeshAsset, dae: collada.Collada, uniq=None,
                   include_skin=True, node_transform: Optional[np.ndarray] = None) -> collada.scene.Node:
    if node_transform is None:
        node_transform = np.identity(4)

    should_skin = False
    skin_seg = ll_mesh.segments.get('skin')
    bind_shape_matrix = None
    if include_skin and skin_seg:
        bind_shape_matrix = llsd_to_mat4(skin_seg["bind_shape_matrix"])
        should_skin = True
        # Transform from the skin will be applied on the controller, not the node
        node_transform = np.identity(4)

    if not uniq:
        uniq = secrets.token_urlsafe(4)

    geom_nodes = []
    node_name = f"mainnode{uniq}"
    # TODO: do the other LODs?
    for submesh_num, submesh in enumerate(ll_mesh.segments["high_lod"]):
        # Make sure none of our IDs collide with those of other nodes
        sub_uniq = uniq + str(submesh_num)

        range_xyz = positions_from_domain(submesh["Position"], submesh["PositionDomain"])
        xyz = np.array([x.data() for x in range_xyz])

        range_uv = positions_from_domain(submesh['TexCoord0'], submesh['TexCoord0Domain'])
        uv = np.array([x.data() for x in range_uv]).flatten()

        norms = np.array([x.data() for x in submesh["Normal"]])

        effect = collada.material.Effect(
            id=f"effect{sub_uniq}",
            params=[],
            specular=(0.0, 0.0, 0.0, 0.0),
            reflectivity=(0.0, 0.0, 0.0, 0.0),
            emission=(0.0, 0.0, 0.0, 0.0),
            ambient=(0.0, 0.0, 0.0, 0.0),
            reflective=0.0,
            shadingtype="blinn",
            shininess=0.0,
            diffuse=(1.0, 1.0, 1.0),
        )
        mat = collada.material.Material(f"material{sub_uniq}", f"material{sub_uniq}", effect)

        dae.materials.append(mat)
        dae.effects.append(effect)

        vert_src = collada.source.FloatSource(f"verts-array{sub_uniq}", xyz.flatten(), ("X", "Y", "Z"))
        norm_src = collada.source.FloatSource(f"norms-array{sub_uniq}", norms.flatten(), ("X", "Y", "Z"))
        # UV maps have to have the same name or they'll behave weirdly when objects are merged.
        uv_src = collada.source.FloatSource("uvs-array", np.array(uv), ("U", "V"))

        geom = collada.geometry.Geometry(dae, f"geometry{sub_uniq}", "geometry", [vert_src, norm_src, uv_src])

        input_list = collada.source.InputList()
        input_list.addInput(0, 'VERTEX', f'#verts-array{sub_uniq}', set="0")
        input_list.addInput(0, 'NORMAL', f'#norms-array{sub_uniq}', set="0")
        input_list.addInput(0, 'TEXCOORD', '#uvs-array', set="0")

        tri_idxs = np.array(submesh["TriangleList"]).flatten()
        matnode = collada.scene.MaterialNode(f"materialref{sub_uniq}", mat, inputs=[])
        tri_set = geom.createTriangleSet(tri_idxs, input_list, f'materialref{sub_uniq}')
        geom.primitives.append(tri_set)
        dae.geometries.append(geom)

        if should_skin:
            joint_names = np.array(skin_seg['joint_names'], dtype=object)
            joints_source = collada.source.NameSource(f"joint-names{sub_uniq}", joint_names, ("JOINT",))
            # PyCollada has a bug where it doesn't set the source URI correctly. Fix it.
            accessor = joints_source.xmlnode.find(f"{dae.tag('technique_common')}/{dae.tag('accessor')}")
            if not accessor.get('source').startswith('#'):
                accessor.set('source', f"#{accessor.get('source')}")

            flattened_bind_poses = []
            for bind_pose in skin_seg['inverse_bind_matrix']:
                flattened_bind_poses.append(mat4_to_collada(llsd_to_mat4(bind_pose)))
            flattened_bind_poses = np.array(flattened_bind_poses)
            inv_bind_source = _create_mat4_source(f"bind-poses{sub_uniq}", flattened_bind_poses, "TRANSFORM")

            weight_joint_idxs = []
            weights = []
            vert_weight_counts = []
            cur_weight_idx = 0
            for vert_weights in submesh['Weights']:
                vert_weight_counts.append(len(vert_weights))
                for vert_weight in vert_weights:
                    weights.append(vert_weight.weight)
                    weight_joint_idxs.append(vert_weight.joint_idx)
                    weight_joint_idxs.append(cur_weight_idx)
                    cur_weight_idx += 1

            weights_source = collada.source.FloatSource(f"skin-weights{sub_uniq}", np.array(weights), ("WEIGHT",))
            # We need to make a controller for each material since materials are essentially distinct meshes
            # in SL, with their own distinct sets of weights and vertex data.
            controller_node = E.controller(
                E.skin(
                    E.bind_shape_matrix(' '.join(str(x) for x in mat4_to_collada(bind_shape_matrix))),
                    joints_source.xmlnode,
                    inv_bind_source.xmlnode,
                    weights_source.xmlnode,
                    E.joints(
                        E.input(semantic="JOINT", source=f"#joint-names{sub_uniq}"),
                        E.input(semantic="INV_BIND_MATRIX", source=f"#bind-poses{sub_uniq}")
                    ),
                    E.vertex_weights(
                        E.input(semantic="JOINT", source=f"#joint-names{sub_uniq}", offset="0"),
                        E.input(semantic="WEIGHT", source=f"#skin-weights{sub_uniq}", offset="1"),
                        E.vcount(' '.join(str(x) for x in vert_weight_counts)),
                        E.v(' '.join(str(x) for x in weight_joint_idxs)),
                        count=str(len(submesh['Weights']))
                    ),
                    source=f"#geometry{sub_uniq}"
                ),
                id=f"Armature-{sub_uniq}",
                name=node_name
            )
            controller = collada.controller.Controller.load(dae, {}, controller_node)
            dae.controllers.append(controller)
            geom_node = collada.scene.ControllerNode(controller, [matnode])
        else:
            geom_node = collada.scene.GeometryNode(geom, [matnode])

        geom_nodes.append(geom_node)

    node = collada.scene.Node(
        node_name,
        children=geom_nodes,
        transforms=[collada.scene.MatrixTransform(mat4_to_collada(node_transform))],
    )
    if should_skin:
        # We need a skeleton per _mesh asset_ because you could have incongruous skeletons
        # within the same linkset.
        # TODO: can we maintain some kind of skeleton cache, where if this skeleton has no conflicts
        #  with another skeleton in the cache, we just use that skeleton and add any additional joints?
        skel_root = load_skeleton_nodes()
        transform_skeleton(skel_root, dae, skin_seg)
        skel = collada.scene.Node.load(dae, skel_root, {})
        skel.children.append(node)
        skel.id = f"Skel-{uniq}"
        skel.save()
        node = skel
    return node


def load_skeleton_nodes() -> etree.ElementBase:
    # TODO: this sucks. Can't we construct nodes with the appropriate transformation
    #  matrices from the data in `avatar_skeleton.xml`?
    skel_path = get_resource_filename("lib/base/data/male_collada_joints.xml")
    with open(skel_path, 'r') as f:
        return etree.fromstring(f.read())


def transform_skeleton(skel_root: etree.ElementBase, dae: collada.Collada, skin_seg: SkinSegmentDict,
                       include_unreferenced_bones=False):
    """Update skeleton XML nodes to account for joint translations in the mesh"""
    joint_nodes: Dict[str, collada.scene.Node] = {}
    for skel_node in skel_root.iter():
        # xpath is loathsome so this is easier.
        if skel_node.tag != dae.tag('node') or skel_node.get('type') != 'JOINT':
            continue
        joint_nodes[skel_node.get('name')] = collada.scene.Node.load(dae, skel_node, {})
    for joint_name, matrix in zip(skin_seg['joint_names'], skin_seg.get('alt_inverse_bind_matrix', [])):
        joint_node = joint_nodes[joint_name]
        joint_decomp = transformations.decompose_matrix(llsd_to_mat4(matrix))
        joint_node.matrix = mat4_to_collada(transformations.compose_matrix(translate=joint_decomp[3]))
        # Update the underlying XML element with the new transform matrix
        joint_node.save()

    if not include_unreferenced_bones:
        needed_heirarchy = set()
        for skel_node in joint_nodes.values():
            skel_node = skel_node.xmlnode
            if skel_node.get('name') in skin_seg['joint_names']:
                # Add this joint and any ancestors the list of needed joints
                while skel_node is not None:
                    needed_heirarchy.add(skel_node.get('name'))
                    skel_node = skel_node.getparent()

        for skel_node in joint_nodes.values():
            skel_node = skel_node.xmlnode
            if skel_node.get('name') not in needed_heirarchy:
                skel_node.getparent().remove(skel_node)

    pelvis_offset = skin_seg.get('pelvis_offset')

    # TODO: should we even do this here? It's not present in the collada, just
    #  something that's specified in the uploader before conversion to LLMesh.
    if pelvis_offset and 'mPelvis' in joint_nodes:
        pelvis_node = joint_nodes['mPelvis']
        # Column-major!
        pelvis_node.matrix[3][2] += pelvis_offset
        pelvis_node.save()


def _create_mat4_source(name: str, data: np.ndarray, semantic: str):
    # PyCollada has no way to make a source with a float4x4 semantic. Do it a bad way.
    # Note that collada demands column-major matrices whereas LLSD mesh has them row-major!
    source = collada.source.FloatSource(name, data, tuple(f"M{x}" for x in range(16)))
    accessor = source.xmlnode[1][0]
    for child in list(accessor):
        accessor.remove(child)
    accessor.append(E.param(name=semantic, type="float4x4"))
    return source


def fix_weird_bind_matrices(skin_seg: SkinSegmentDict) -> None:
    """
    Fix weird-looking bind matrices to have sensible scaling and rotations

    Sometimes we get enormous inverse bind matrices (each component 10k+) and tiny
    bind shape matrix components. This detects inverse bind shape matrices
    with weird scales and tries to set them to what they "should" be without
    the weird inverted scaling.
    """

    # Sometimes we get mesh assets that have the vertex data naturally in y-up orientation,
    # and get re-oriented to z-up not through the bind shape matrix, but through the
    # transforms in the inverse bind matrices!
    #
    # Blender, for one, does not like this very much, and generally won't generate mesh
    # assets like this, as explained here https://developer.blender.org/T38660.
    # In vanilla Blender, these mesh assets will show up scaled and rotated _only_ according
    # to the bind shape matrix, which may end up with the model 25 meters tall and sitting
    # on its side.
    #
    # https://avalab.org/avastar/292/knowledge/compare-workbench/, while somewhat outdated,
    # has some information on rest pose vs default pose and scaling that I believe is relevant.
    # https://github.com/KhronosGroup/glTF-Blender-IO/issues/994 as well.
    #
    # While trying to figure out what was going on, I searched for something like
    # "inverse bind matrix scale collada", "bind pose scale blender", etc. Pretty much every
    # result was either a bug filed by, or a question asked by the creator of Avastar, or an SL user.
    # I think that says a lot about how annoying it is to author mesh for SL in particular.
    #
    # I spent a good month or so tearing my hair out over this wondering how these values could
    # even be possible. I wasn't sure how I should write mesh import code if I don't understand
    # how to interpret existing data, or how it even ended up the way it did. Turns out I wasn't
    # misinterpreting the data, the data really is just weird.
    #
    # I'd also had the idea that you could sniff which body a given rigged asset was meant
    # for by doing trivial matching on the inverse bind matrices, but obviously that isn't true!
    #
    # Basically:
    # 1) Maya is evil and generates evil, this evil bleeds into SL's assets through transforms.
    # 2) Blender is also evil, but in a manner that doesn't agree with Maya's evil.
    # 3) Collada was a valiant effort, but is evil in practice. Seemingly simple Collada
    #    files are interpreted completely differently by Blender, Maya, and sometimes SL.
    # 4) Those three evils collude to make an interop nightmare for everyone like "oh my rigger
    #    rigs using Maya and now my model is huge and all my normals are fucked on reimport"
    # 5) Yes, there's still good reasons to be using Avastar in 2022 even though nobody authoring
    #    rigged mesh for any other use has to use something similar.

    if not skin_seg['joint_names']:
        return

    # TODO: calculate the correct inverse bind matrix scale & rotations from avatar_skeleton.xml
    #  definitions. If the rotation and scale factors are the same across all inverse bind matrices then
    #  they can be moved over to the bind shape matrix to keep Blender happy.
    #  Maybe add a scaled / rotated empty as a parent for the armature instead?
    return


def main():
    # Take an llmesh file as an argument and spit out basename-converted.dae
    with open(sys.argv[1], "rb") as f:
        reader = BufferReader("<", f.read())

    mesh = mesh_to_collada(reader.read(LLMeshSerializer(parse_segment_contents=True)))
    mesh.write(sys.argv[1].rsplit(".", 1)[0] + "-converted.dae")


if __name__ == "__main__":
    main()
